# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import subprocess
import time
from uuid import uuid4

import boto.ec2

from toil.lib.aws import zone_to_region
from toil.lib.retry import retry
from toil.provisioners.aws import get_best_aws_zone
from toil.test import ToilTest, needs_aws_ec2, needs_fetchable_appliance

log = logging.getLogger(__name__)

@needs_aws_ec2
@needs_fetchable_appliance
class AbstractClusterTest(ToilTest):
    def __init__(self, methodName):
        super().__init__(methodName=methodName)
        self.keyName = os.getenv('TOIL_AWS_KEYNAME').strip() or 'id_rsa'
        self.clusterName = 'aws-provisioner-test-' + str(uuid4())
        self.leaderNodeType = 't2.medium'
        self.clusterType = 'mesos'
        self.zone = get_best_aws_zone()
        assert self.zone is not None, "Could not determine AWS availability zone to test in; is TOIL_AWS_ZONE set?"
        # We need a boto2 connection to EC2 to check on the cluster
        self.boto2_ec2 = boto.ec2.connect_to_region(zone_to_region(self.zone))
        # Where should we put our virtualenv?
        self.venvDir = '/tmp/venv'

    def python(self):
        """
        Return the full path to the venv Python on the leader.
        """
        return os.path.join(self.venvDir, 'bin/python')

    def pip(self):
        """
        Return the full path to the venv pip on the leader.
        """
        return os.path.join(self.venvDir, 'bin/pip')

    def destroyCluster(self) -> None:
        """
        Destroy the cluster we built, if it exists.

        Succeeds if the cluster does not currently exist.
        """
        subprocess.check_call(['toil', 'destroy-cluster', '-p=aws', '-z', self.zone, self.clusterName])

    def setUp(self):
        """
        Set up for the test.
        Must be overridden to call this method and set self.jobStore.
        """
        super().setUp()
        # Make sure that destroy works before we create any clusters.
        # If this fails, no tests will run.
        self.destroyCluster()

    def tearDown(self):
        # Note that teardown will run even if the test crashes.
        super().tearDown()
        self.destroyCluster()
        subprocess.check_call(['toil', 'clean', self.jobStore])

    def sshUtil(self, command):
        """
        Run the given command on the cluster.
        Raise subprocess.CalledProcessError if it fails.
        """

        cmd = ['toil', 'ssh-cluster', '--insecure', '-p=aws', '-z', self.zone, self.clusterName] + command
        log.info("Running %s.", str(cmd))
        p = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        # Put in non-blocking mode. See https://stackoverflow.com/a/59291466
        os.set_blocking(p.stdout.fileno(), False)
        os.set_blocking(p.stderr.fileno(), False)

        out_buffer = b''
        err_buffer = b''

        loops_since_line = 0

        running = True
        while running:
            # While the process is running, see if it stopped
            running = (p.poll() is None)

            # Also collect its output
            out_data = p.stdout.read()
            if out_data:
                out_buffer += out_data

            while out_buffer.find(b'\n') != -1:
                # And log every full line
                cut = out_buffer.find(b'\n')
                log.info('STDOUT: %s', out_buffer[0:cut].decode('utf-8', errors='ignore'))
                loops_since_line = 0
                out_buffer = out_buffer[cut+1:]

            # Same for the error
            err_data = p.stderr.read()
            if err_data:
                err_buffer += err_data

            while err_buffer.find(b'\n') != -1:
                cut = err_buffer.find(b'\n')
                log.info('STDERR: %s', err_buffer[0:cut].decode('utf-8', errors='ignore'))
                loops_since_line = 0
                err_buffer = err_buffer[cut+1:]

            loops_since_line += 1
            if loops_since_line > 60:
                log.debug('...waiting...')
                loops_since_line = 0

            time.sleep(1)

        # At the end, log the last lines
        if out_buffer:
            log.info('STDOUT: %s', out_buffer.decode('utf-8', errors='ignore'))
        if err_buffer:
            log.info('STDERR: %s', err_buffer.decode('utf-8', errors='ignore'))

        if p.returncode != 0:
            # It failed
            log.error("Failed to run %s.", str(cmd))
            raise subprocess.CalledProcessError(p.returncode, ' '.join(cmd))

    @retry(errors=[subprocess.CalledProcessError], intervals=[1, 1])
    def createClusterUtil(self, args=None):
        args = [] if args is None else args

        command = ['toil', 'launch-cluster', '-p=aws', '-z', self.zone, f'--keyPairName={self.keyName}',
                   f'--leaderNodeType={self.leaderNodeType}', f'--clusterType={self.clusterType}', '--logDebug', self.clusterName] + args

        log.debug('Launching cluster: %s', command)

        # Try creating the cluster
        subprocess.check_call(command)
        # If we fail, tearDown will destroy the cluster.

    def launchCluster(self):
        self.createClusterUtil()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import codecs
import logging
import os
import random
import sys
import time
import traceback
from threading import Event, Thread
from unittest import skipIf

import pytest

from toil.batchSystems.singleMachine import SingleMachineBatchSystem
from toil.job import Job
from toil.leader import DeadlockException
from toil.exceptions import FailedJobsException
from toil.lib.retry import retry_flaky_test
from toil.test import ToilTest, get_temp_file, slow

logger = logging.getLogger(__name__)


class JobServiceTest(ToilTest):
    """
    Tests testing the Job.Service class
    """

    @slow
    def testServiceSerialization(self):
        """
        Tests that a service can receive a promise without producing a serialization
        error.
        """
        job = Job()
        service = ToySerializableService("woot")
        startValue = job.addService(service) # Add a first service to job
        subService = ToySerializableService(startValue) # Now create a child of
        # that service that takes the start value promise from the parent service
        job.addService(subService, parentService=service) # This should work if
        # serialization on services is working correctly.

        self.runToil(job)

    @slow
    def testService(self, checkpoint=False):
        """
        Tests the creation of a Job.Service with random failures of the worker.
        """
        for test in range(2):
            outFile = get_temp_file(rootDir=self._createTempDir()) # Temporary file
            messageInt = random.randint(1, sys.maxsize)
            try:
                # Wire up the services/jobs
                t = Job.wrapJobFn(serviceTest, outFile, messageInt, checkpoint=checkpoint)

                # Run the workflow repeatedly until success
                self.runToil(t)

                # Check output
                self.assertEqual(int(open(outFile).readline()), messageInt)
            finally:
                os.remove(outFile)

    @slow
    @skipIf(SingleMachineBatchSystem.numCores < 4, 'Need at least four cores to run this test')
    def testServiceDeadlock(self):
        """
        Creates a job with more services than maxServices, checks that deadlock is detected.
        """
        outFile = get_temp_file(rootDir=self._createTempDir())
        try:
            def makeWorkflow():
                job = Job()
                r1 = job.addService(ToySerializableService("woot1"))
                r2 = job.addService(ToySerializableService("woot2"))
                r3 = job.addService(ToySerializableService("woot3"))
                job.addChildFn(fnTest, [ r1, r2, r3 ], outFile)
                return job

            # This should fail as too few services available
            try:
                self.runToil(makeWorkflow(), badWorker=0.0, maxServiceJobs=2, deadlockWait=5)
            except DeadlockException:
                print("Got expected deadlock exception")
            else:
                assert 0

            # This should pass, as adequate services available
            self.runToil(makeWorkflow(), maxServiceJobs=3)
            # Check we get expected output
            assert open(outFile).read() == "woot1 woot2 woot3"
        finally:
            os.remove(outFile)

    def testServiceWithCheckpoints(self):
        """
        Tests the creation of a Job.Service with random failures of the worker, making the root job use checkpointing to
        restart the subtree.
        """
        self.testService(checkpoint=True)

    @slow
    @skipIf(SingleMachineBatchSystem.numCores < 4, 'Need at least four cores to run this test')
    def testServiceRecursive(self, checkpoint=True):
        """
        Tests the creation of a Job.Service, creating a chain of services and accessing jobs.
        Randomly fails the worker.
        """
        for test in range(1):
            # Temporary file
            outFile = get_temp_file(rootDir=self._createTempDir())
            messages = [ random.randint(1, sys.maxsize) for i in range(3) ]
            try:
                # Wire up the services/jobs
                t = Job.wrapJobFn(serviceTestRecursive, outFile, messages, checkpoint=checkpoint)

                # Run the workflow repeatedly until success
                self.runToil(t)

                # Check output
                self.assertEqual(list(map(int, open(outFile).readlines())), messages)
            finally:
                os.remove(outFile)

    @slow
    @skipIf(SingleMachineBatchSystem.numCores < 4, 'Need at least four cores to run this test')
    @retry_flaky_test(prepare=[ToilTest.tearDown, ToilTest.setUp])
    @pytest.mark.timeout(1200)
    def testServiceParallelRecursive(self, checkpoint=True):
        """
        Tests the creation of a Job.Service, creating parallel chains of services and accessing jobs.
        Randomly fails the worker.
        """
        for test in range(1):
            # Temporary file
            outFiles = [get_temp_file(rootDir=self._createTempDir()) for j in range(2)]
            messageBundles = [ [ random.randint(1, sys.maxsize) for i in range(3) ] for j in range(2) ]
            try:
                # Wire up the services/jobs
                t = Job.wrapJobFn(serviceTestParallelRecursive, outFiles, messageBundles, checkpoint=True)

                # Run the workflow repeatedly until success
                self.runToil(t, retryCount=2)

                # Check output
                for (messages, outFile) in zip(messageBundles, outFiles):
                    self.assertEqual(list(map(int, open(outFile).readlines())), messages)
            finally:
                list(map(os.remove, outFiles))

    def runToil(self, rootJob, retryCount=1, badWorker=0.5, badWorkedFailInterval=0.1, maxServiceJobs=sys.maxsize, deadlockWait=60):
        # Create the runner for the workflow.
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.logLevel = "DEBUG"

        options.retryCount = retryCount
        options.badWorker = badWorker
        options.badWorkerFailInterval = badWorkedFailInterval
        options.servicePollingInterval = 1
        options.maxServiceJobs = maxServiceJobs
        options.deadlockWait=deadlockWait

        # Run the workflow
        totalTrys = 0
        while True:
            try:
                Job.Runner.startToil(rootJob, options)
                break
            except FailedJobsException as e:
                i = e.numberOfFailedJobs
                if totalTrys > 50: #p(fail after this many restarts) = 0.5**32
                    self.fail() #Exceeded a reasonable number of restarts
                totalTrys += 1
                options.restart = True

class PerfectServiceTest(JobServiceTest):
    def runToil(self, rootJob, retryCount=1, badWorker=0, badWorkedFailInterval=1000, maxServiceJobs=sys.maxsize, deadlockWait=60):
        """
        Let us run all the tests in the other service test class, but without worker failures.
        """
        super().runToil(rootJob, retryCount, badWorker, badWorkedFailInterval, maxServiceJobs, deadlockWait)


def serviceTest(job, outFile, messageInt):
    """
    Creates one service and one accessing job, which communicate with two files to establish
    that both run concurrently.
    """
    #Clean out out-file
    open(outFile, 'w').close()
    randInt = random.randint(1, sys.maxsize) # We create a random number that is added to messageInt and subtracted by the serviceAccessor, to prove that
    # when service test is checkpointed and restarted there is never a connection made between an earlier service and later serviceAccessor, or vice versa.
    job.addChildJobFn(serviceAccessor, job.addService(ToyService(messageInt + randInt)), outFile, randInt)

def serviceTestRecursive(job, outFile, messages):
    """
    Creates a chain of services and accessing jobs, each paired together.
    """
    if len(messages) > 0:
        #Clean out out-file
        open(outFile, 'w').close()
        randInt = random.randint(1, sys.maxsize)
        service = ToyService(messages[0] + randInt)
        child = job.addChildJobFn(serviceAccessor, job.addService(service), outFile, randInt)

        for i in range(1, len(messages)):
            randInt = random.randint(1, sys.maxsize)
            service2 = ToyService(messages[i] + randInt, cores=0.1)
            child = child.addChildJobFn(serviceAccessor,
                                        job.addService(service2, parentService=service),
                                        outFile, randInt, cores=0.1)
            service = service2

def serviceTestParallelRecursive(job, outFiles, messageBundles):
    """
    Creates multiple chains of services and accessing jobs.
    """
    for messages, outFile in zip(messageBundles, outFiles):
        #Clean out out-file
        open(outFile, 'w').close()
        if len(messages) > 0:
            randInt = random.randint(1, sys.maxsize)
            service = ToyService(messages[0] + randInt)
            child = job.addChildJobFn(serviceAccessor, job.addService(service), outFile, randInt)

            for i in range(1, len(messages)):
                randInt = random.randint(1, sys.maxsize)
                service2 = ToyService(messages[i] + randInt, cores=0.1)
                child = child.addChildJobFn(serviceAccessor,
                                            job.addService(service2, parentService=service),
                                            outFile, randInt, cores=0.1)
                service = service2

class ToyService(Job.Service):
    def __init__(self, messageInt, *args, **kwargs):
        """
        While established the service repeatedly:
             - reads an integer i from the inJobStoreFileID file
             - writes i and messageInt to the outJobStoreFileID file
        """
        Job.Service.__init__(self, *args, **kwargs)
        self.messageInt = messageInt

    def start(self, job):
        assert self.disk is not None
        assert self.memory is not None
        assert self.cores is not None
        self.terminate = Event()
        self.error = Event()
        # Note that service jobs are special and do not necessarily have job.jobStoreID.
        # So we don't associate these files with this job.
        inJobStoreID = job.fileStore.jobStore.get_empty_file_store_id()
        outJobStoreID = job.fileStore.jobStore.get_empty_file_store_id()
        self.serviceThread = Thread(target=self.serviceWorker,
                                    args=(job.fileStore.jobStore, self.terminate, self.error,
                                          inJobStoreID, outJobStoreID,
                                          self.messageInt))
        self.serviceThread.start()
        return (inJobStoreID, outJobStoreID)

    def stop(self, job):
        self.terminate.set()
        self.serviceThread.join()

    def check(self):
        if self.error.isSet():
            raise RuntimeError("Service worker failed")
        return True

    @staticmethod
    def serviceWorker(jobStore, terminate, error, inJobStoreID, outJobStoreID, messageInt):
        try:
            while True:
                if terminate.isSet(): # Quit if we've got the terminate signal
                    logger.debug("Service worker being told to quit")
                    return

                time.sleep(0.2) # Sleep to avoid thrashing

                # Try reading a line from the input file
                try:
                    with jobStore.read_file_stream(inJobStoreID) as f:
                        f = codecs.getreader('utf-8')(f)
                        line = f.readline()
                except:
                    logger.debug("Something went wrong reading a line: %s", traceback.format_exc())
                    raise

                if len(line.strip()) == 0:
                    # Don't try and make an integer out of nothing
                    continue

                # Try converting the input line into an integer
                try:
                    inputInt = int(line)
                except ValueError:
                    logger.debug("Tried casting input line '%s' to integer but got error: %s", line, traceback.format_exc())
                    continue

                # Write out the resulting read integer and the message
                with jobStore.update_file_stream(outJobStoreID) as f:
                    f.write((f"{inputInt} {messageInt}\n").encode())
        except:
            logger.debug("Error in service worker: %s", traceback.format_exc())
            error.set()
            raise

def serviceAccessor(job, communicationFiles, outFile, randInt):
    """
    Writes a random integer iinto the inJobStoreFileID file, then tries 10 times reading
    from outJobStoreFileID to get a pair of integers, the first equal to i the second written into the outputFile.
    """
    inJobStoreFileID, outJobStoreFileID = communicationFiles

    # Get a random integer
    key = random.randint(1, sys.maxsize)

    # Write the integer into the file
    logger.debug("Writing key to inJobStoreFileID")
    with job.fileStore.jobStore.update_file_stream(inJobStoreFileID) as fH:
        fH.write(("%s\n" % key).encode('utf-8'))

    logger.debug("Trying to read key and message from outJobStoreFileID")
    for i in range(10): # Try 10 times over
        time.sleep(0.2) #Avoid thrashing

        # Try reading an integer from the input file and writing out the message
        with job.fileStore.jobStore.read_file_stream(outJobStoreFileID) as fH:
            fH = codecs.getreader('utf-8')(fH)
            line = fH.readline()

        tokens = line.split()
        if len(tokens) != 2:
            continue

        key2, message = tokens

        if int(key2) == key:
            logger.debug(f"Matched key's: {key}, writing message: {int(message) - randInt} with randInt: {randInt}")
            with open(outFile, 'a') as fH:
                fH.write("%s\n" % (int(message) - randInt))
            return

    assert 0 # Job failed to get info from the service

class ToySerializableService(Job.Service):
    def __init__(self, messageInt, *args, **kwargs):
        """
        Trivial service for testing serialization.
        """
        Job.Service.__init__(self, *args, **kwargs)
        self.messageInt = messageInt

    def start(self, job):
        return self.messageInt

    def stop(self, job):
        pass

    def check(self):
        return True

def fnTest(strings, outputFile):
    """
    Function concatenates the strings together and writes them to the output file
    """
    with open(outputFile, 'w') as fH:
        fH.write(" ".join(strings))



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import bz2
import logging
import os
import types
from ssl import SSLError
from typing import Optional, cast

from boto3.s3.transfer import TransferConfig
from boto.exception import BotoServerError, S3ResponseError, SDBResponseError
from botocore.client import Config
from botocore.exceptions import ClientError
from mypy_boto3_s3 import S3Client, S3ServiceResource

from toil.lib.aws import session
from toil.lib.aws.utils import connection_reset, get_bucket_region
from toil.lib.compatibility import compat_bytes
from toil.lib.retry import (DEFAULT_DELAYS,
                            DEFAULT_TIMEOUT,
                            ErrorCondition,
                            get_error_code,
                            get_error_message,
                            get_error_status,
                            old_retry,
                            retry)

logger = logging.getLogger(__name__)

# Make one botocore Config object for setting up S3 to talk to the exact region
# we told it to talk to, so that we don't keep making new objects that look
# like distinct memoization keys. We need to set the addressing style to path
# so that we talk to region-specific DNS names and not per-bucket DNS names. We
# also need to set a special flag to make sure we don't use the generic
# s3.amazonaws.com for us-east-1, or else we might not actually end up talking
# to us-east-1 when a bucket is there.
DIAL_SPECIFIC_REGION_CONFIG = Config(s3={
    'addressing_style': 'path',
    'us_east_1_regional_endpoint': 'regional'
})

class SDBHelper:
    """
    A mixin with methods for storing limited amounts of binary data in an SDB item

    >>> import os
    >>> H=SDBHelper
    >>> H.presenceIndicator() # doctest: +ALLOW_UNICODE
    u'numChunks'
    >>> H.binaryToAttributes(None)['numChunks']
    0
    >>> H.attributesToBinary({u'numChunks': 0})
    (None, 0)
    >>> H.binaryToAttributes(b'') # doctest: +ALLOW_UNICODE +ALLOW_BYTES
    {u'000': b'VQ==', u'numChunks': 1}
    >>> H.attributesToBinary({u'numChunks': 1, u'000': b'VQ=='}) # doctest: +ALLOW_BYTES
    (b'', 1)

    Good pseudo-random data is very likely smaller than its bzip2ed form. Subtract 1 for the type
    character, i.e  'C' or 'U', with which the string is prefixed. We should get one full chunk:

    >>> s = os.urandom(H.maxRawValueSize-1)
    >>> d = H.binaryToAttributes(s)
    >>> len(d), len(d['000'])
    (2, 1024)
    >>> H.attributesToBinary(d) == (s, 1)
    True

    One byte more and we should overflow four bytes into the second chunk, two bytes for
    base64-encoding the additional character and two bytes for base64-padding to the next quartet.

    >>> s += s[0:1]
    >>> d = H.binaryToAttributes(s)
    >>> len(d), len(d['000']), len(d['001'])
    (3, 1024, 4)
    >>> H.attributesToBinary(d) == (s, 2)
    True

    """
    # The SDB documentation is not clear as to whether the attribute value size limit of 1024
    # applies to the base64-encoded value or the raw value. It suggests that responses are
    # automatically encoded from which I conclude that the limit should apply to the raw,
    # unencoded value. However, there seems to be a discrepancy between how Boto computes the
    # request signature if a value contains a binary data, and how SDB does it. This causes
    # requests to fail signature verification, resulting in a 403. We therefore have to
    # base64-encode values ourselves even if that means we loose a quarter of capacity.

    maxAttributesPerItem = 256
    maxValueSize = 1024
    maxRawValueSize = maxValueSize * 3 // 4
    # Just make sure we don't have a problem with padding or integer truncation:
    assert len(base64.b64encode(b' ' * maxRawValueSize)) == 1024
    assert len(base64.b64encode(b' ' * (1 + maxRawValueSize))) > 1024

    @classmethod
    def _reservedAttributes(cls):
        """
        Override in subclass to reserve a certain number of attributes that can't be used for
        chunks.
        """
        return 1

    @classmethod
    def _maxChunks(cls):
        return cls.maxAttributesPerItem - cls._reservedAttributes()

    @classmethod
    def maxBinarySize(cls, extraReservedChunks=0):
        return (cls._maxChunks() - extraReservedChunks) * cls.maxRawValueSize - 1  # for the 'C' or 'U' prefix

    @classmethod
    def _maxEncodedSize(cls):
        return cls._maxChunks() * cls.maxValueSize

    @classmethod
    def binaryToAttributes(cls, binary):
        """
        Turn a bytestring, or None, into SimpleDB attributes.
        """
        if binary is None: return {'numChunks': 0}
        assert isinstance(binary, bytes)
        assert len(binary) <= cls.maxBinarySize()
        # The use of compression is just an optimization. We can't include it in the maxValueSize
        # computation because the compression ratio depends on the input.
        compressed = bz2.compress(binary)
        if len(compressed) > len(binary):
            compressed = b'U' + binary
        else:
            compressed = b'C' + compressed
        encoded = base64.b64encode(compressed)
        assert len(encoded) <= cls._maxEncodedSize()
        n = cls.maxValueSize
        chunks = (encoded[i:i + n] for i in range(0, len(encoded), n))
        attributes = {cls._chunkName(i): chunk for i, chunk in enumerate(chunks)}
        attributes.update({'numChunks': len(attributes)})
        return attributes

    @classmethod
    def _chunkName(cls, i):
        return str(i).zfill(3)

    @classmethod
    def _isValidChunkName(cls, s):
        return len(s) == 3 and s.isdigit()

    @classmethod
    def presenceIndicator(cls):
        """
        The key that is guaranteed to be present in the return value of binaryToAttributes().
        Assuming that binaryToAttributes() is used with SDB's PutAttributes, the return value of
        this method could be used to detect the presence/absence of an item in SDB.
        """
        return 'numChunks'

    @classmethod
    def attributesToBinary(cls, attributes):
        """
        :rtype: (str|None,int)
        :return: the binary data and the number of chunks it was composed from
        """
        chunks = [(int(k), v) for k, v in attributes.items() if cls._isValidChunkName(k)]
        chunks.sort()
        numChunks = int(attributes['numChunks'])
        if numChunks:
            serializedJob = b''.join(v.encode() for k, v in chunks)
            compressed = base64.b64decode(serializedJob)
            if compressed[0] == b'C'[0]:
                binary = bz2.decompress(compressed[1:])
            elif compressed[0] == b'U'[0]:
                binary = compressed[1:]
            else:
                raise RuntimeError(f'Unexpected prefix {compressed[0]}')
        else:
            binary = None
        return binary, numChunks


def fileSizeAndTime(localFilePath):
    file_stat = os.stat(localFilePath)
    return file_stat.st_size, file_stat.st_mtime


@retry(errors=[ErrorCondition(
    error=ClientError,
    error_codes=[404, 500, 502, 503, 504]
)])
def uploadFromPath(localFilePath: str,
                   resource,
                   bucketName: str,
                   fileID: str,
                   headerArgs: Optional[dict] = None,
                   partSize: int = 50 << 20):
    """
    Uploads a file to s3, using multipart uploading if applicable

    :param str localFilePath: Path of the file to upload to s3
    :param S3.Resource resource: boto3 resource
    :param str bucketName: name of the bucket to upload to
    :param str fileID: the name of the file to upload to
    :param dict headerArgs: http headers to use when uploading - generally used for encryption purposes
    :param int partSize: max size of each part in the multipart upload, in bytes

    :return: version of the newly uploaded file
    """
    if headerArgs is None:
        headerArgs = {}

    client = resource.meta.client
    file_size, file_time = fileSizeAndTime(localFilePath)

    version = uploadFile(localFilePath, resource, bucketName, fileID, headerArgs, partSize)
    info = client.head_object(Bucket=bucketName, Key=compat_bytes(fileID), VersionId=version, **headerArgs)
    size = info.get('ContentLength')

    assert size == file_size

    # Make reasonably sure that the file wasn't touched during the upload
    assert fileSizeAndTime(localFilePath) == (file_size, file_time)
    return version


@retry(errors=[ErrorCondition(
    error=ClientError,
    error_codes=[404, 500, 502, 503, 504]
)])
def uploadFile(readable,
               resource,
               bucketName: str,
               fileID: str,
               headerArgs: Optional[dict] = None,
               partSize: int = 50 << 20):
    """
    Upload a readable object to s3, using multipart uploading if applicable.
    :param readable: a readable stream or a file path to upload to s3
    :param S3.Resource resource: boto3 resource
    :param str bucketName: name of the bucket to upload to
    :param str fileID: the name of the file to upload to
    :param dict headerArgs: http headers to use when uploading - generally used for encryption purposes
    :param int partSize: max size of each part in the multipart upload, in bytes
    :return: version of the newly uploaded file
    """
    if headerArgs is None:
        headerArgs = {}

    client = resource.meta.client
    config = TransferConfig(
        multipart_threshold=partSize,
        multipart_chunksize=partSize,
        use_threads=True
    )
    if isinstance(readable, str):
        client.upload_file(Filename=readable,
                           Bucket=bucketName,
                           Key=compat_bytes(fileID),
                           ExtraArgs=headerArgs,
                           Config=config)
    else:
        client.upload_fileobj(Fileobj=readable,
                              Bucket=bucketName,
                              Key=compat_bytes(fileID),
                              ExtraArgs=headerArgs,
                              Config=config)

        # Wait until the object exists before calling head_object
        object_summary = resource.ObjectSummary(bucketName, compat_bytes(fileID))
        object_summary.wait_until_exists(**headerArgs)

    info = client.head_object(Bucket=bucketName, Key=compat_bytes(fileID), **headerArgs)
    return info.get('VersionId', None)

class ServerSideCopyProhibitedError(RuntimeError):
    """
    Raised when AWS refuses to perform a server-side copy between S3 keys, and
    insists that you pay to download and upload the data yourself instead.
    """
    pass

@retry(errors=[ErrorCondition(
    error=ClientError,
    error_codes=[404, 500, 502, 503, 504]
)])
def copyKeyMultipart(resource: S3ServiceResource,
                     srcBucketName: str,
                     srcKeyName: str,
                     srcKeyVersion: str,
                     dstBucketName: str,
                     dstKeyName: str,
                     sseAlgorithm: Optional[str] = None,
                     sseKey: Optional[str] = None,
                     copySourceSseAlgorithm: Optional[str] = None,
                     copySourceSseKey: Optional[str] = None):
    """
    Copies a key from a source key to a destination key in multiple parts. Note that if the
    destination key exists it will be overwritten implicitly, and if it does not exist a new
    key will be created. If the destination bucket does not exist an error will be raised.

    This function will always do a fast, server-side copy, at least
    until/unless <https://github.com/boto/boto3/issues/3270> is fixed. In some
    situations, a fast, server-side copy is not actually possible. For example,
    when residing in an AWS VPC with an S3 VPC Endpoint configured, copying
    from a bucket in another region to a bucket in your own region cannot be
    performed server-side. This is because the VPC Endpoint S3 API servers
    refuse to perform server-side copies between regions, the source region's
    API servers refuse to initiate the copy and refer you to the destination
    bucket's region's API servers, and the VPC routing tables are configured to
    redirect all access to the current region's S3 API servers to the S3
    Endpoint API servers instead.

    If a fast server-side copy is not actually possible, a
    ServerSideCopyProhibitedError will be raised.

    :param resource: boto3 resource
    :param str srcBucketName: The name of the bucket to be copied from.
    :param str srcKeyName: The name of the key to be copied from.
    :param str srcKeyVersion: The version of the key to be copied from.
    :param str dstBucketName: The name of the destination bucket for the copy.
    :param str dstKeyName: The name of the destination key that will be created or overwritten.
    :param str sseAlgorithm: Server-side encryption algorithm for the destination.
    :param str sseKey: Server-side encryption key for the destination.
    :param str copySourceSseAlgorithm: Server-side encryption algorithm for the source.
    :param str copySourceSseKey: Server-side encryption key for the source.

    :rtype: str
    :return: The version of the copied file (or None if versioning is not enabled for dstBucket).
    """
    dstBucket = resource.Bucket(compat_bytes(dstBucketName))
    dstObject = dstBucket.Object(compat_bytes(dstKeyName))
    copySource = {'Bucket': compat_bytes(srcBucketName), 'Key': compat_bytes(srcKeyName)}
    if srcKeyVersion is not None:
        copySource['VersionId'] = compat_bytes(srcKeyVersion)

    # Get a client to the source region, which may not be the same as the one
    # this resource is connected to. We should probably talk to it for source
    # object metadata. And we really want it to talk to the source region and
    # not wherever the bucket virtual hostnames go.
    source_region = get_bucket_region(srcBucketName)
    source_client = cast(
        S3Client,
        session.client(
            's3',
            region_name=source_region,
            config=DIAL_SPECIFIC_REGION_CONFIG
        )
    )

    # The boto3 functions don't allow passing parameters as None to
    # indicate they weren't provided. So we have to do a bit of work
    # to ensure we only provide the parameters when they are actually
    # required.
    destEncryptionArgs = {}
    if sseKey is not None:
        destEncryptionArgs.update({'SSECustomerAlgorithm': sseAlgorithm,
                                   'SSECustomerKey': sseKey})
    copyEncryptionArgs = {}
    if copySourceSseKey is not None:
        copyEncryptionArgs.update({'CopySourceSSECustomerAlgorithm': copySourceSseAlgorithm,
                                   'CopySourceSSECustomerKey': copySourceSseKey})
    copyEncryptionArgs.update(destEncryptionArgs)

    try:
        # Kick off a server-side copy operation
        dstObject.copy(copySource, SourceClient=source_client, ExtraArgs=copyEncryptionArgs)
    except ClientError as e:
        if get_error_code(e) == 'AccessDenied' and 'cross-region' in get_error_message(e):
            # We have this problem: <https://aws.amazon.com/premiumsupport/knowledge-center/s3-troubleshoot-copy-between-buckets/#Cross-Region_request_issues_with_VPC_endpoints_for_Amazon_S3>
            # The Internet and AWS docs say that we just can't do a
            # cross-region CopyObject from inside a VPC with an endpoint. The
            # AGC team suggested we try doing the copy by talking to the
            # source region instead, but that does not actually work; if we
            # hack up botocore enough for it to actually send the request to
            # the source region's API servers, they reject it and tell us to
            # talk to the destination region's API servers instead. Which we
            # can't reach.
            logger.error('Amazon is refusing to perform a server-side copy of %s: %s', copySource, e)
            raise ServerSideCopyProhibitedError()
        else:
            # Some other ClientError happened
            raise


    # Wait until the object exists before calling head_object
    object_summary = resource.ObjectSummary(dstObject.bucket_name, dstObject.key)
    object_summary.wait_until_exists(**destEncryptionArgs)

    # Unfortunately, boto3's managed copy doesn't return the version
    # that it actually copied to. So we have to check immediately
    # after, leaving open the possibility that it may have been
    # modified again in the few seconds since the copy finished. There
    # isn't much we can do about it.
    info = resource.meta.client.head_object(Bucket=dstObject.bucket_name,
                                            Key=dstObject.key,
                                            **destEncryptionArgs)
    return info.get('VersionId', None)


def _put_attributes_using_post(self, domain_or_name, item_name, attributes,
                               replace=True, expected_value=None):
    """
    Monkey-patched version of SDBConnection.put_attributes that uses POST instead of GET

    The GET version is subject to the URL length limit which kicks in before the 256 x 1024 limit
    for attribute values. Using POST prevents that.

    https://github.com/BD2KGenomics/toil/issues/502
    """
    domain, domain_name = self.get_domain_and_name(domain_or_name)
    params = {'DomainName': domain_name,
              'ItemName': item_name}
    self._build_name_value_list(params, attributes, replace)
    if expected_value:
        self._build_expected_value(params, expected_value)
    # The addition of the verb keyword argument is the only difference to put_attributes (Hannes)
    return self.get_status('PutAttributes', params, verb='POST')


def monkeyPatchSdbConnection(sdb):
    """
    :type sdb: SDBConnection
    """
    sdb.put_attributes = types.MethodType(_put_attributes_using_post, sdb)

def sdb_unavailable(e):
    # Since we're checking against a collection here we absolutely need an
    # integer status code. This is probably a BotoServerError, but other 500s
    # and 503s probably ought to be retried too.
    return get_error_status(e) in (500, 503)


def no_such_sdb_domain(e):
    return (isinstance(e, SDBResponseError)
            and e.error_code
            and e.error_code.endswith('NoSuchDomain'))


def retryable_ssl_error(e):
    # https://github.com/BD2KGenomics/toil/issues/978
    return isinstance(e, SSLError) and e.reason == 'DECRYPTION_FAILED_OR_BAD_RECORD_MAC'


def retryable_sdb_errors(e):
    return (sdb_unavailable(e)
            or no_such_sdb_domain(e)
            or connection_reset(e)
            or retryable_ssl_error(e))


def retry_sdb(delays=DEFAULT_DELAYS, timeout=DEFAULT_TIMEOUT, predicate=retryable_sdb_errors):
    return old_retry(delays=delays, timeout=timeout, predicate=predicate)




baseVersion = '5.11.0'
cgcloudVersion = '1.6.0a1.dev393'
version = '5.11.0-9a04dabb36d6ab13ed1ac7c711dbdc8c71724dc9'
distVersion = '5.11.0'
exactPython = 'python3.9'
python = 'python3.9'
dockerTag = '5.11.0-9a04dabb36d6ab13ed1ac7c711dbdc8c71724dc9-py3.9'
currentCommit = '9a04dabb36d6ab13ed1ac7c711dbdc8c71724dc9'
dockerRegistry = 'quay.io/ucsc_cgl'
dockerName = 'toil'
dirty = False
cwltool_version = '3.1.20230425144158'



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Message types and message bus for leader component coordination.

Historically, the Toil Leader has been organized around functions calling other
functions to "handle" different things happening. Over time, it has become very
brittle: exactly the right handling functions need to be called in exactly the
right order, or it gets confused and does the wrong thing.

The MessageBus is meant to let the leader avoid this by more losely coupling
its components together, by having them communicate by sending messages instead
of by calling functions.

When events occur (like a job coming back from the batch system with a failed
exit status), this will be translated into a message that will be sent to the
bus. Then, all the leader components that need to react to this message in some
way (by, say, decrementing the retry count) would listen for the relevant
messages on the bus and react to them. If a new component needs to be added, it
can be plugged into the message bus and receive and react to messages without
interfering with existing components' ability to react to the same messages.

Eventually, the different aspects of the Leader could become separate objects.

By default, messages stay entirely within the Toil leader process, and are not
persisted anywhere, not even in the JobStore.

The Message Bus also provides an extension point: its messages can be
serialized to a file by the leader (see the --writeMessages option), and they
can then be decoded using MessageBus.scan_bus_messages() (as is done in the
Toil WES server backend). By replaying the messages and tracking their effects
on job state, you can get an up-to-date view of the state of the jobs in a
workflow. This includes information, such as whether jobs are issued or
running, or what jobs have completely finished, which is not persisted in the
JobStore.

The MessageBus instance for the leader process is owned by the Toil leader, but
the BatchSystem has an opportunity to connect to it, and can send (or listen
for) messages. Right now the BatchSystem deos not *have* to send or receive any
messages; the Leader is responsible for polling it via the BatchSystem API and
generating the events. But a BatchSystem implementation *may* send additional
events (like JobAnnotationMessage).

Currently, the MessageBus is implemented using pypubsub, and so messages are
always handled in a single Thread, the Toil leader's main loop thread. If other
components send events, they will be shipped over to that thread inside the
MessageBus. Communication between processes is allowed using
MessageBus.connect_output_file() and MessageBus.scan_bus_messages().
"""

import collections
from dataclasses import dataclass
import inspect
import logging

import os
import queue
import json
import tempfile
import threading
from typing import (IO,
                    Any,
                    Callable,
                    Dict,
                    Iterator,
                    List,
                    NamedTuple,
                    Optional,
                    Type,
                    TypeVar,
                    Union,
                    cast)

from pubsub.core import Publisher
from pubsub.core.listener import Listener
from pubsub.core.topicobj import Topic
from pubsub.core.topicutils import ALL_TOPICS

logger = logging.getLogger( __name__ )

# We define a bunch of named tuple message types.
# These all need to be plain data: only hold ints, strings, etc.

class JobIssuedMessage(NamedTuple):
    """
    Produced when a job is issued to run on the batch system.
    """
    # The kind of job issued, for statistics aggregation
    job_type: str
    # The job store ID of the job
    job_id: str
    # The toil batch ID of the job
    toil_batch_id: int

class JobUpdatedMessage(NamedTuple):
    """
    Produced when a job is "updated" and ready to have something happen to it.
    """
    # The job store ID of the job
    job_id: str
    # The error code/return code for the job, which is nonzero if something has
    # gone wrong, and 0 otherwise.
    result_status: int

class JobCompletedMessage(NamedTuple):
    """
    Produced when a job is completed, whether successful or not.
    """
    # The kind of job issued, for statistics aggregation
    job_type: str
    # The job store ID of the job
    job_id: str
    # Exit code for job_id
    exit_code: int

class JobFailedMessage(NamedTuple):
    """
    Produced when a job is completely failed, and will not be retried again.
    """
    # The kind of job issued, for statistics aggregation
    job_type: str
    # The job store ID of the job
    job_id: str

class JobMissingMessage(NamedTuple):
    """
    Produced when a job goes missing and should be in the batch system but isn't.
    """
    # The job store ID of the job
    job_id: str

class JobAnnotationMessage(NamedTuple):
    """
    Produced when extra information (such as an AWS Batch job ID from the
    AWSBatchBatchSystem) is available that goes with a job.
    """
    # The job store ID of the job
    job_id: str
    # The name of the annotation
    annotation_name: str
    # The annotation data
    annotation_value: str

class ExternalBatchIdMessage(NamedTuple):
    """
    Produced when using a batch system, links toil assigned batch ID to
    Batch system ID (Whatever's returned by local implementation, PID, batch ID, etc)
    """
    #Assigned toil batch job id
    toil_batch_id: int
    #Batch system scheduler identity
    external_batch_id: str
    #Batch system name
    batch_system: str

class QueueSizeMessage(NamedTuple):
    """
    Produced to describe the size of the queue of jobs issued but not yet
    completed. Theoretically recoverable from other messages.
    """
    # The size of the queue
    queue_size: int

class ClusterSizeMessage(NamedTuple):
    """
    Produced by the Toil-integrated autoscaler describe the number of
    instances of a certain type in a cluster.
    """
    # The instance type name, like t4g.medium
    instance_type: str
    # The number of instances of that type that the Toil autoscaler thinks
    # there are
    current_size: int

class ClusterDesiredSizeMessage(NamedTuple):
    """
    Produced by the Toil-integrated autoscaler to describe the number of
    instances of a certain type that it thinks will be needed.
    """
    # The instance type name, like t4g.medium
    instance_type: str
    # The number of instances of that type that the Toil autoscaler wants there
    # to be
    desired_size: int

# Then we define a serialization format.

def message_to_bytes(message: NamedTuple) -> bytes:
    """
    Convert a plain-old-data named tuple into a byte string.
    """
    parts = []
    for item in message:
        if isinstance(item, (int, float, bool)) or item is None:
            # This also handles e.g. values from an IntEnum, where the type extends int.
            # They might replace __str__() but we hope they use a compatible __format__()
            parts.append(f"{item}".encode('utf-8'))
        elif isinstance(item, str):
            parts.append(item.encode('unicode_escape'))
        else:
            # We haven't implemented this type yet.
            raise RuntimeError(f"Cannot store message argument of type {type(item)}: {item}")
    return b'\t'.join(parts)


# TODO: Messages have to be named tuple types.
MessageType = TypeVar('MessageType')
def bytes_to_message(message_type: Type[MessageType], data: bytes) -> MessageType:
    """
    Convert bytes from message_to_bytes back to a message of the given type.
    """
    parts = data.split(b'\t')

    # Get a mapping from field name to type in the named tuple.
    # We need to check a couple different fields because this moved in a recent
    # Python 3 release.
    field_to_type: Optional[Dict[str, type]] = cast(Optional[Dict[str, type]],
                                                    getattr(message_type, '__annotations__',
                                                            getattr(message_type, '_field_types', None)))
    if field_to_type is None:
        raise RuntimeError(f"Cannot get field types from {message_type}")
    field_names: List[str] = getattr(message_type, '_fields')

    if len(field_names) != len(parts):
        raise RuntimeError(f"Cannot parse {field_names} from {parts}")

    # Parse each part according to its type and put it in here
    typed_parts = []

    for name, part in zip(field_names, parts):
        field_type = field_to_type[name]
        if field_type in [int, float, bool]:
            typed_parts.append(field_type(part.decode('utf-8')))
        elif field_type == str:
            # Decode, accounting for escape sequences
            typed_parts.append(part.decode('unicode_escape'))
        else:
            raise RuntimeError(f"Cannot read message argument of type {field_type}")

    # Build the parts back up into the named tuple.
    return message_type(*typed_parts)




class MessageBus:
    """
    Holds messages that should cause jobs to change their scheduling states.
    Messages are put in and buffered, and can be taken out and handled as
    batches when convenient.

    All messages are NamedTuple objects of various subtypes.

    Message order is guaranteed to be preserved within a type.
    """

    def __init__(self) -> None:
        # Each MessageBus has an independent PyPubSub instance
        self._pubsub = Publisher()

        # We will deliver messages to the thread we were constructed in
        self._owning_thread = threading.get_ident()

        # Messages from other threads go in this queue
        self._queue: queue.Queue[Any] = queue.Queue()

    @classmethod
    def _type_to_name(cls, message_type: type) -> str:
        """
        Convert a type to a name that can be a PyPubSub topic (all normal
        characters, hierarchically dotted).
        """

        return '.'.join([message_type.__module__, message_type.__name__])

    # All our messages are NamedTuples, but NamedTuples don't actually inherit
    # from NamedTupe, so MyPy complains if we require that here.
    def publish(self, message: Any) -> None:
        """
        Put a message onto the bus. Can be called from any thread.
        """

        if threading.get_ident() == self._owning_thread:
            # We can deliver this now
            self._deliver(message)
            # Also do any waiting messages
            self.check()
        else:
            # We need to put it in the queue for the main thread
            self._queue.put(message)

    def check(self) -> None:
        """
        If we are in the owning thread, deliver any messages that are in the
        queue for us. Must be called every once in a while in the main thread,
        possibly through inbox objects.
        """

        if threading.get_ident() == self._owning_thread:
            # We are supposed to receive messages
            while True:
                # Until we can't get a message, get one
                message: Optional[Any] = None
                try:
                    message = self._queue.get_nowait()
                except queue.Empty:
                    # No messages are waiting
                    return
                # Handle the message
                self._deliver(message)

    def _deliver(self, message: Any) -> None:
        """
        Runs only in the owning thread. Delivers a message to its listeners.
        """
        topic = self._type_to_name(type(message))
        logger.debug('Notifying %s with message: %s', topic, message)
        self._pubsub.sendMessage(topic, message=message)

    # This next function takes callables that take things of the type that was passed in as a
    # runtime argument, which we can explain to MyPy using a TypeVar and Type[]
    MessageType = TypeVar('MessageType', bound='NamedTuple')
    def subscribe(self, message_type: Type[MessageType], handler: Callable[[MessageType], Any]) -> Listener:
        """
        Register the given callable to be called when messages of the given type are sent.
        It will be called with messages sent after the subscription is created.
        Returns a subscription object; when the subscription object is GC'd the subscription will end.
        """

        topic = self._type_to_name(message_type)
        logger.debug('Listening for message topic: %s', topic)

        # Make sure to wrap the handler so we get the right argument name and
        # we can control lifetime.
        def handler_wraper(message: MessageBus.MessageType) -> None:
            handler(message)

        # The docs says this returns the listener but really it seems to return
        # a listener and something else.
        listener, _ = self._pubsub.subscribe(handler_wraper, topic)
        # Hide the handler function in the pubsub listener to keep it alive.
        # If it goes out of scope the subscription expires, and the pubsub
        # system only uses weak references.
        setattr(listener, 'handler_wrapper', handler_wraper)
        return listener

    def connect(self, wanted_types: List[type]) -> 'MessageBusConnection':
        """
        Get a connection object that serves as an inbox for messages of the
        given types.
        Messages of those types will accumulate in the inbox until it is
        destroyed. You can check for them at any time.
        """
        connection = MessageBusConnection()
        # We call this private method, really we mean this to be module scope.
        connection._set_bus_and_message_types(self, wanted_types)
        return connection

    def outbox(self) -> 'MessageOutbox':
        """
        Get a connection object that only allows sending messages.
        """
        connection = MessageOutbox()
        connection._set_bus(self)
        return connection

    def connect_output_file(self, file_path: str) -> Any:
        """
        Send copies of all messages to the given output file.

        Returns connection data which must be kept alive for the
        connection to persist. That data is opaque: the user is not supposed to
        look at it or touch it or do anything with it other than store it
        somewhere or delete it.
        """


        stream = open(file_path, 'wb')

        # Type of the ** is the value type of the dictionary; key type is always string.
        def handler(topic_object: Topic = Listener.AUTO_TOPIC, **message_data: NamedTuple) -> None:
            """
            Log the message in the given message data, associated with the
            given topic.
            """
            # There should always be a "message"
            assert len(message_data) == 1
            assert 'message' in message_data
            message = message_data['message']
            topic = topic_object.getName()
            stream.write(topic.encode('utf-8'))
            stream.write(b'\t')
            stream.write(message_to_bytes(message))
            stream.write(b'\n')
            stream.flush()

        listener, _ = self._pubsub.subscribe(handler, ALL_TOPICS)

        # To keep this alive, we need to keep the handler alive, and we might
        # want the pypubsub Listener.
        return (handler, listener)


    # TODO: If we annotate this as returning an Iterator[NamedTuple], MyPy
    # complains when we loop over it that the loop variable is a <nothing>,
    # ifen in code protected by isinstance(). Using a typevar makes it complain
    # that we "need type annotation" for the loop variable, because it can't
    # get it from the types in the (possibly empty?) list.
    # TODO: Find a good way to annotate this as returning an iterator of a
    # union of the types passed in message_types, in a way that MyPy can
    # understand.
    @classmethod
    def scan_bus_messages(cls, stream: IO[bytes], message_types: List[Type[NamedTuple]]) -> Iterator[Any]:
        """
        Get an iterator over all messages in the given log stream of the given
        types, in order. Discard any trailing partial messages.
        """

        # We know all the message types we care about, so build the mapping from topic name to message type
        name_to_type = {cls._type_to_name(t): t for t in message_types}

        for line in stream:
            logger.debug('Got message: %s', line)
            if not line.endswith(b'\n'):
                # Skip unterminated line
                continue
            # Drop the newline and split on first tab
            parts = line[:-1].split(b'\t', 1)

            # Get the type of the message
            message_type = name_to_type.get(parts[0].decode('utf-8'))
            if message_type is None:
                # We aren't interested in this kind of message.
                continue

            # Decode the actual message
            message = bytes_to_message(message_type, parts[1])

            # And produce it
            yield message

class MessageBusClient:
    """
    Base class for clients (inboxes and outboxes) of a message bus. Handles
    keeping a reference to the message bus.
    """

    def __init__(self) -> None:
        """
        Make a disconnected client.
        """

        # We might be given a reference to the message bus
        self._bus: Optional[MessageBus] = None

    def _set_bus(self, bus: MessageBus) -> None:
        """
        Connect to the given bus.

        Should only ever be called once on a given instance.
        """
        self._bus = bus

class MessageInbox(MessageBusClient):
    """
    A buffered connection to a message bus that lets us receive messages.
    Buffers incoming messages until you are ready for them.
    Does not preserve ordering between messages of different types.
    """

    def __init__(self) -> None:
        """
        Make a disconnected inbox.
        """

        super().__init__()

        # This holds all the messages on the bus, organized by type.
        self._messages_by_type: Dict[type, List[Any]] = {}
        # This holds listeners for all the types, when we connect to a bus
        self._listeners_by_type: Dict[type, Listener] = {}

        # We define a handler for messages
        def on_message(message: Any) -> None:
            self._messages_by_type[type(message)].append(message)
        self._handler = on_message

    def _set_bus_and_message_types(self, bus: MessageBus, wanted_types: List[type]) -> None:
        """
        Connect to the given bus and collect the given message types.

        Should only ever be called once on a given instance.
        """

        self._set_bus(bus)

        for t in wanted_types:
            # For every kind of message we are subscribing to

            # Make a queue for the messages
            self._messages_by_type[t] = []
            # Make and save a subscription
            self._listeners_by_type[t] = bus.subscribe(t, self._handler)

    def _check_bus(self) -> None:
        """
        Make sure we are connected to a bus, and let the bus check for messages
        and deliver them to us. We call this before we access our delivered
        messages.
        """

        if self._bus is None:
            raise RuntimeError("Cannot receive message when not connected to a bus")
        # Make sure to check for messages form other threads first
        self._bus.check()

    def count(self, message_type: type) -> int:
        """
        Get the number of pending messages of the given type.
        """
        self._check_bus()
        return len(self._messages_by_type[message_type])

    def empty(self) -> bool:
        """
        Return True if no messages are pending, and false otherwise.
        """
        self._check_bus()
        return all(len(v) == 0 for v in self._messages_by_type.values())

    # This next function returns things of the type that was passed in as a
    # runtime argument, which we can explain to MyPy using a TypeVar and Type[]
    MessageType = TypeVar('MessageType')
    def for_each(self, message_type: Type[MessageType]) -> Iterator[MessageType]:
        """
        Loop over all messages currently pending of the given type. Each that
        is handled without raising an exception will be removed.

        Messages sent while this function is running will not be yielded by the
        current call.
        """

        self._check_bus()

        # Grab the message buffer for this kind of message.
        message_list = self._messages_by_type[message_type]
        # Make a new buffer. TODO: Will be hard to be thread-safe because other
        # threads could have a reference to the old buffer.
        self._messages_by_type[message_type] = []

        # Flip around to chronological order
        message_list.reverse()
        try:
            while len(message_list) > 0:
                # We need to handle the case where a new message of this type comes
                # in while we're looping, from the handler. So we take each off the
                # list from the end while we handle it, and put it back if
                # something goes wrong.
                message = message_list.pop()
                handled = False
                try:
                    # Emit the message
                    assert isinstance(message, message_type), f"Unacceptable message type {type(message)} in list for type {message_type}"
                    yield message
                    # If we get here it was handled without error.
                    handled = True
                finally:
                    if not handled:
                        # An exception happened, and we're bailing out of the
                        # while loop. Make sure the message isn't dropped, in
                        # case someone wants to recover and handle it again
                        # later with another for_each call.
                        message_list.append(message)
        finally:
            # Dump anything remaining in our buffer back into the main buffer,
            # in the right order, and before the later messages.
            message_list.reverse()
            self._messages_by_type[message_type] = message_list + self._messages_by_type[message_type]

class MessageOutbox(MessageBusClient):
    """
    A connection to a message bus that lets us publish messages.
    """

    def __init__(self) -> None:
        """
        Make a disconnected outbox.
        """
        super().__init__()

    def publish(self, message: Any) -> None:
        """
        Publish the given message to the connected message bus.

        We have this so you don't need to store both the bus and your connection.
        """
        if self._bus is None:
            raise RuntimeError("Cannot send message when not connected to a bus")
        self._bus.publish(message)

class MessageBusConnection(MessageInbox, MessageOutbox):
    """
    A two-way connection to a message bus. Buffers incoming messages until you
    are ready for them, and lets you send messages.
    """

    def __init__(self) -> None:
        """
        Make a MessageBusConnection that is not connected yet.
        """
        super().__init__()

    def _set_bus_and_message_types(self, bus: MessageBus, wanted_types: List[type]) -> None:
        """
        Connect to the given bus and collect the given message types.

        We must not have connected to anything yet.
        """

        # We need to call the two inherited connection methods
        super()._set_bus_and_message_types(bus, wanted_types)
        self._set_bus(bus)


@dataclass
class JobStatus:
    """
    Records the status of a job.
    """

    job_store_id: str
    name: str
    exit_code: int
    annotations: Dict[str, str]
    toil_batch_id: int
    external_batch_id: str
    batch_system: str

    def __repr__(self) -> str:
        return json.dumps(self, default= lambda o: o.__dict__, indent=4)
def replay_message_bus(path: str) -> Dict[str, JobStatus]:
    """
    Replay all the messages and work out what they mean for jobs.

    We track the state and name of jobs here, by ID.
    We would use a list of two items but MyPy can't understand a list
    of items of multiple types, so we need to define a new class.

    Returns a dictionary from the job_id to a dataclass, JobStatus.
    A JobStatus contains information about a job which we have gathered
    from the message bus, including the job store id, name of the job
    the exit code, any associated annotations, the toil batch id
    the external batch id, and the batch system on which the job
    is running.
    """

    job_statuses: Dict[str, JobStatus] = collections.defaultdict(lambda: JobStatus('', '', -1, {}, -1, '', ''))
    batch_to_job_id = {}
    try:
        with open(path, 'rb') as log_stream:
            # Read all the full, properly-terminated messages about job updates
            for event in MessageBus.scan_bus_messages(log_stream, [JobUpdatedMessage, JobIssuedMessage, JobCompletedMessage,
                                                                   JobFailedMessage, JobAnnotationMessage, ExternalBatchIdMessage]):
                # And for each of them
                logger.info('Got message from workflow: %s', event)

                if isinstance(event, JobUpdatedMessage):
                    # Apply the latest return code from the job with this ID.
                    job_statuses[event.job_id].exit_code = event.result_status
                elif isinstance(event, JobIssuedMessage):
                    job_statuses[event.job_id].job_store_id = event.job_id
                    job_statuses[event.job_id].name = event.job_type
                    job_statuses[event.job_id].toil_batch_id = event.toil_batch_id
                    job_statuses[event.job_id].exit_code = -1
                    batch_to_job_id[event.toil_batch_id] = event.job_id
                elif isinstance(event, JobCompletedMessage):
                    job_statuses[event.job_id].name = event.job_type
                    job_statuses[event.job_id].exit_code = event.exit_code
                elif isinstance(event, JobFailedMessage):
                    job_statuses[event.job_id].name = event.job_type
                    if job_statuses[event.job_id].exit_code == 0:
                        # Record the failure if we never got a failed exit code.
                        job_statuses[event.job_id].exit_code = 1
                elif isinstance(event, JobAnnotationMessage):
                    # Remember the last value of any annotation that is set
                    job_statuses[event.job_id].annotations[event.annotation_name] = event.annotation_value
                elif isinstance(event, ExternalBatchIdMessage):
                    if event.toil_batch_id in batch_to_job_id:
                        job_statuses[batch_to_job_id[event.toil_batch_id]].external_batch_id = event.external_batch_id
                        job_statuses[batch_to_job_id[event.toil_batch_id]].batch_system = event.batch_system
    except FileNotFoundError:
        logger.warning("We were unable to access the file")

    return job_statuses

def gen_message_bus_path() -> str:
    """
    Return a file path in tmp to store the message bus at.
    Calling function is responsible for cleaning the generated file.
    """
    fd, path = tempfile.mkstemp()
    os.close(fd)
    return path
    #TODO Might want to clean up the tmpfile at some point after running the workflow



# Modified from: https://github.com/common-workflow-language/workflow-service
import functools
import json
import logging
import os
import tempfile
from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import urldefrag

import connexion  # type: ignore
from werkzeug.utils import secure_filename

logger = logging.getLogger(__name__)

# Define a type for WES task log entries in responses
# TODO: make this a typed dict with all the WES task log field names and their types.
TaskLog = Dict[str, Union[str, int, None]]


class VersionNotImplementedException(Exception):
    """
    Raised when the requested workflow version is not implemented.
    """
    def __init__(self,
                 wf_type: str, version: Optional[str] = None, supported_versions: Optional[List[str]] = None) -> None:
        if version:
            message = ("workflow_type '{}' requires 'workflow_type_version' to be one of '{}'.  "
                       "Got '{}' instead.".format(wf_type, str(supported_versions), version))
        else:
            message = f"workflow_type '{wf_type}' is not supported."

        super().__init__(message)


class MalformedRequestException(Exception):
    """
    Raised when the request is malformed.
    """
    def __init__(self, message: str) -> None:
        super().__init__(message)


class WorkflowNotFoundException(Exception):
    """
    Raised when the requested run ID is not found.
    """
    def __init__(self) -> None:
        super().__init__("The requested workflow run wasn't found.")


class WorkflowConflictException(Exception):
    """
    Raised when the requested workflow is not in the expected state.
    """
    def __init__(self, run_id: str):
        super().__init__(f"Workflow {run_id} exists when it shouldn't.")


class OperationForbidden(Exception):
    """
    Raised when the request is forbidden.
    """
    def __init__(self, message: str) -> None:
        super().__init__(message)


class WorkflowExecutionException(Exception):
    """
    Raised when an internal error occurred during the execution of the workflow.
    """
    def __init__(self, message: str) -> None:
        super().__init__(message)


def handle_errors(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    This decorator catches errors from the wrapped function and returns a JSON
    formatted error message with the appropriate status code defined by the
    GA4GH WES spec.
    """

    def error(msg: Any, code: int = 500) -> Tuple[Dict[str, Any], int]:
        logger.warning(f"Exception raised when calling '{func.__name__}()':", exc_info=True)
        return {"msg": str(msg), "status_code": code}, code

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except MalformedRequestException as e:
            return error(e, code=400)
        except VersionNotImplementedException as e:
            return error(e, code=400)
        except OperationForbidden as e:
            return error(e, code=403)
        except (FileNotFoundError, WorkflowNotFoundException) as e:
            return error(e, code=404)
        except WorkflowConflictException as e:
            return error(e, code=400)
        except WorkflowExecutionException as e:
            return error(e, code=500)
        except Exception as e:
            return error(e, code=500)

    return wrapper


class WESBackend:
    """
    A class to represent a GA4GH Workflow Execution Service (WES) API backend.
    Intended to be inherited. Subclasses should implement all abstract methods
    to handle user requests when they hit different endpoints.
    """

    def __init__(self, options: List[str]):
        """
        :param options: A list of default engine options to use when executing
                        a workflow.  Example options:
                        ["--logLevel=CRITICAL","--workDir=/path/to/dir",
                            "--tag=Name=default", "--tag=Owner=shared", ...]
        """
        self.options = options or []

    def resolve_operation_id(self, operation_id: str) -> Any:
        """
        Map an operationId defined in the OpenAPI or swagger yaml file to a
        function.

        :param operation_id: The operation ID defined in the specification.
        :returns: A function that should be called when the given endpoint is
                  reached.
        """
        return getattr(self, operation_id.split(".")[-1])

    @abstractmethod
    def get_service_info(self) -> Dict[str, Any]:
        """
        Get information about the Workflow Execution Service.

        GET /service-info
        """
        raise NotImplementedError

    @abstractmethod
    def list_runs(self, page_size: Optional[int] = None, page_token: Optional[str] = None) -> Dict[str, Any]:
        """
        List the workflow runs.

        GET /runs
        """
        raise NotImplementedError

    @abstractmethod
    def run_workflow(self) -> Dict[str, str]:
        """
        Run a workflow. This endpoint creates a new workflow run and returns
        a `RunId` to monitor its progress.

        POST /runs
        """
        raise NotImplementedError

    @abstractmethod
    def get_run_log(self, run_id: str) -> Dict[str, Any]:
        """
        Get detailed info about a workflow run.

        GET /runs/{run_id}
        """
        raise NotImplementedError

    @abstractmethod
    def cancel_run(self, run_id: str) -> Dict[str, str]:
        """
        Cancel a running workflow.

        POST /runs/{run_id}/cancel
        """
        raise NotImplementedError

    @abstractmethod
    def get_run_status(self, run_id: str) -> Dict[str, str]:
        """
        Get quick status info about a workflow run, returning a simple result
        with the overall state of the workflow run.

        GET /runs/{run_id}/status
        """
        raise NotImplementedError

    @staticmethod
    def log_for_run(run_id: Optional[str], message: str) -> None:
        if run_id:
            logging.info("Workflow %s: %s", run_id, message)
        else:
            logging.info(message)

    @staticmethod
    def secure_path(path: str) -> str:
        return os.path.join(*[str(secure_filename(p)) for p in path.split("/") if p not in ("", ".", "..")])

    def collect_attachments(self, run_id: Optional[str], temp_dir: Optional[str]) -> Tuple[str, Dict[str, Any]]:
        """
        Collect attachments from the current request by staging uploaded files
        to temp_dir, and return the temp_dir and parsed body of the request.

        :param run_id: The run ID for logging.
        :param temp_dir: The directory where uploaded files should be staged.
                         If None, a temporary directory is created.
        """
        if not temp_dir:
            temp_dir = tempfile.mkdtemp()
        body: Dict[str, Any] = {}
        has_attachments = False
        for key, ls in connexion.request.files.lists():
            try:
                for value in ls:
                    # uploaded files that are required to execute the workflow
                    if key == "workflow_attachment":
                        # guard against maliciously constructed filenames
                        dest = os.path.join(temp_dir, self.secure_path(value.filename))
                        if not os.path.isdir(os.path.dirname(dest)):
                            os.makedirs(os.path.dirname(dest))
                        self.log_for_run(run_id, f"Staging attachment '{value.filename}' to '{dest}'")
                        value.save(dest)
                        has_attachments = True
                        body[key] = f"file://{temp_dir}"  # Reference to temp working dir.

                    elif key in ("workflow_params", "tags", "workflow_engine_parameters"):
                        content = value.read()
                        body[key] = json.loads(content.decode("utf-8"))
                    else:
                        body[key] = value.read().decode()
            except Exception as e:
                raise MalformedRequestException(f"Error reading parameter '{key}': {e}")

        for key, ls in connexion.request.form.lists():
            try:
                for value in ls:
                    if not value:
                        continue
                    if key in ("workflow_params", "tags", "workflow_engine_parameters"):
                        body[key] = json.loads(value)
                    else:
                        body[key] = value
            except Exception as e:
                raise MalformedRequestException(f"Error reading parameter '{key}': {e}")

        if "workflow_url" in body:
            url, ref = urldefrag(body["workflow_url"])
            if ":" not in url:
                if not has_attachments:
                    raise MalformedRequestException("Relative 'workflow_url' but missing 'workflow_attachment'")
                body["workflow_url"] = self.secure_path(url)  # keep this relative
                if ref:
                    # append "#ref" after the url
                    body["workflow_url"] += "#" + self.secure_path(ref)
            self.log_for_run(run_id, "Using workflow_url '%s'" % body.get("workflow_url"))
        else:
            raise MalformedRequestException("Missing 'workflow_url' in submission")

        if "workflow_params" in body and not isinstance(body["workflow_params"], dict):
            # They sent us something silly like "workflow_params": "5"
            raise MalformedRequestException("Got a 'workflow_params' which does not decode to a JSON object")

        return temp_dir, body



# Copyright (C) 2021 Michael R. Crusoe
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import uuid
from typing import Optional

from toil.jobStores.aws.jobStore import AWSJobStore
from toil.lib.aws.session import establish_boto3_session
from toil.lib.aws.utils import create_s3_bucket, get_bucket_region
from toil.test import ToilTest, needs_aws_s3

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


@needs_aws_s3
class S3Test(ToilTest):
    """Confirm the workarounds for us-east-1."""

    from mypy_boto3_s3 import S3ServiceResource
    from mypy_boto3_s3.service_resource import Bucket

    s3_resource: Optional[S3ServiceResource]
    bucket: Optional[Bucket]

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        session = establish_boto3_session(region_name="us-east-1")
        cls.s3_resource = session.resource("s3", region_name="us-east-1")
        cls.bucket = None

    def test_create_bucket(self) -> None:
        """Test bucket creation for us-east-1."""
        bucket_name = f"toil-s3test-{uuid.uuid4()}"
        assert self.s3_resource
        S3Test.bucket = create_s3_bucket(self.s3_resource, bucket_name, "us-east-1")
        S3Test.bucket.wait_until_exists()
        owner_tag = os.environ.get("TOIL_OWNER_TAG")
        if owner_tag:
            bucket_tagging = self.s3_resource.BucketTagging(bucket_name)
            bucket_tagging.put(
                Tagging={"TagSet": [{"Key": "Owner", "Value": owner_tag}]}
            )
        self.assertEqual(get_bucket_region(bucket_name), "us-east-1")

        # Make sure all the bucket location getting strategies work on a bucket we created
        self.assertEqual(get_bucket_region(bucket_name, only_strategies = {1}), "us-east-1")
        self.assertEqual(get_bucket_region(bucket_name, only_strategies = {2}), "us-east-1")
        self.assertEqual(get_bucket_region(bucket_name, only_strategies = {3}), "us-east-1")

    def test_get_bucket_location_public_bucket(self) -> None:
        """
        Test getting buket location for a bucket we don't own.
        """

        bucket_name = 'spacenet-dataset'
        # This bucket happens to live in us-east-1
        self.assertEqual(get_bucket_region(bucket_name), "us-east-1")

    @classmethod
    def tearDownClass(cls) -> None:
        if cls.bucket:
            AWSJobStore._delete_bucket(cls.bucket)
        super().tearDownClass()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import random
import time
import uuid
from argparse import Namespace
from collections import defaultdict
from contextlib import contextmanager
from queue import Empty, Queue
from threading import Event, Thread
from typing import Dict, List, Optional, Set, Tuple
from unittest.mock import MagicMock

from toil.batchSystems.abstractBatchSystem import (AbstractBatchSystem,
                                                   AbstractScalableBatchSystem,
                                                   NodeInfo)
from toil.common import Config, defaultTargetTime
from toil.job import JobDescription
from toil.lib.conversions import human2bytes as h2b
from toil.provisioners.abstractProvisioner import AbstractProvisioner, Shape
from toil.provisioners.clusterScaler import (BinPackedFit,
                                             ClusterScaler,
                                             NodeReservation,
                                             ScalerThread)
from toil.provisioners.node import Node
from toil.test import ToilTest, slow

logger = logging.getLogger(__name__)

# simplified c4.8xlarge (preemptible)
c4_8xlarge_preemptible = Shape(wallTime=3600,
                               memory=h2b('60G'),
                               cores=36,
                               disk=h2b('100G'),
                               preemptible=True)
# simplified c4.8xlarge (non-preemptible)
c4_8xlarge = Shape(wallTime=3600,
                   memory=h2b('60G'),
                   cores=36,
                   disk=h2b('100G'),
                   preemptible=False)
# simplified r3.8xlarge (non-preemptible)
r3_8xlarge = Shape(wallTime=3600,
                   memory=h2b('260G'),
                   cores=32,
                   disk=h2b('600G'),
                   preemptible=False)
# simplified r5.2xlarge (non-preemptible)
r5_2xlarge = Shape(wallTime=3600,
                   memory=h2b('64Gi'),
                   cores=8,
                   disk=h2b('50G'),
                   preemptible=False)
# simplified r5.4xlarge (non-preemptible)
r5_4xlarge = Shape(wallTime=3600,
                   memory=h2b('128Gi'),
                   cores=16,
                   disk=h2b('50G'),
                   preemptible=False)
# simplified t2.micro (non-preemptible)
t2_micro = Shape(wallTime=3600,
                 memory=h2b('1G'),
                 cores=1,
                 disk=h2b('8G'),
                 preemptible=False)

class BinPackingTest(ToilTest):
    def setUp(self):
        self.node_shapes_for_testing = [c4_8xlarge_preemptible, r3_8xlarge]
        self.bpf = BinPackedFit(self.node_shapes_for_testing)

    def testPackingOneShape(self):
        """Pack one shape and check that the resulting reservations look sane."""
        self.bpf.nodeReservations[c4_8xlarge_preemptible] = [NodeReservation(c4_8xlarge_preemptible)]
        self.bpf.addJobShape(Shape(wallTime=1000,
                                   cores=2,
                                   memory=h2b('1G'),
                                   disk=h2b('2G'),
                                   preemptible=True))
        self.assertEqual(self.bpf.nodeReservations[r3_8xlarge], [])
        self.assertEqual([x.shapes() for x in self.bpf.nodeReservations[c4_8xlarge_preemptible]],
                         [[Shape(wallTime=1000,
                                 memory=h2b('59G'),
                                 cores=34,
                                 disk=h2b('98G'),
                                 preemptible=True),
                           Shape(wallTime=2600,
                                 memory=h2b('60G'),
                                 cores=36,
                                 disk=h2b('100G'),
                                 preemptible=True)]])

    def testSorting(self):
        """
        Test that sorting is correct: preemptible, then memory, then cores, then disk,
        then wallTime.
        """
        shapeList = [c4_8xlarge_preemptible, r3_8xlarge, c4_8xlarge, c4_8xlarge,
                     t2_micro, t2_micro, c4_8xlarge, r3_8xlarge, r3_8xlarge, t2_micro]
        shapeList.sort()
        assert shapeList == [c4_8xlarge_preemptible,
                             t2_micro, t2_micro, t2_micro,
                             c4_8xlarge, c4_8xlarge, c4_8xlarge,
                             r3_8xlarge, r3_8xlarge, r3_8xlarge]

    def testAddingInitialNode(self):
        """Pack one shape when no nodes are available and confirm that we fit one node properly."""
        self.bpf.addJobShape(Shape(wallTime=1000,
                                   cores=2,
                                   memory=h2b('1G'),
                                   disk=h2b('2G'),
                                   preemptible=True))
        self.assertEqual([x.shapes() for x in self.bpf.nodeReservations[c4_8xlarge_preemptible]],
                         [[Shape(wallTime=1000,
                                 memory=h2b('59G'),
                                 cores=34,
                                 disk=h2b('98G'),
                                 preemptible=True),
                           Shape(wallTime=2600,
                                 memory=h2b('60G'),
                                 cores=36,
                                 disk=h2b('100G'),
                                 preemptible=True)]])

    def testLowTargetTime(self):
        """
        Test that a low targetTime (0) parallelizes jobs aggressively (1000 queued jobs require
        1000 nodes).

        Ideally, low targetTime means: Start quickly and maximize parallelization after the
        cpu/disk/mem have been packed.

        Disk/cpu/mem packing is prioritized first, so we set job resource reqs so that each
        t2.micro (1 cpu/8G disk/1G RAM) can only run one job at a time with its resources.

        Each job is parametrized to take 300 seconds, so (the minimum of) 1 of them should fit into
        each node's 0 second window, so we expect 1000 nodes.
        """
        allocation = self.run1000JobsOnMicros(jobCores=1,
                                              jobMem=h2b('1G'),
                                              jobDisk=h2b('1G'),
                                              jobTime=300,
                                              globalTargetTime=0)
        self.assertEqual(allocation, {t2_micro: 1000})

    def testHighTargetTime(self):
        """
        Test that a high targetTime (3600 seconds) maximizes packing within the targetTime.

        Ideally, high targetTime means: Maximize packing within the targetTime after the
        cpu/disk/mem have been packed.

        Disk/cpu/mem packing is prioritized first, so we set job resource reqs so that each
        t2.micro (1 cpu/8G disk/1G RAM) can only run one job at a time with its resources.

        Each job is parametrized to take 300 seconds, so 12 of them should fit into each node's
        3600 second window.  1000/12 = 83.33, so we expect 84 nodes.
        """
        allocation = self.run1000JobsOnMicros(jobCores=1,
                                              jobMem=h2b('1G'),
                                              jobDisk=h2b('1G'),
                                              jobTime=300,
                                              globalTargetTime=3600)
        self.assertEqual(allocation, {t2_micro: 84})

    def testZeroResourceJobs(self):
        """
        Test that jobs requiring zero cpu/disk/mem pack first, regardless of targetTime.

        Disk/cpu/mem packing is prioritized first, so we set job resource reqs so that each
        t2.micro (1 cpu/8G disk/1G RAM) can run a seemingly infinite number of jobs with its
        resources.

        Since all jobs should pack cpu/disk/mem-wise on a t2.micro, we expect only one t2.micro to
        be provisioned.  If we raise this, as in testLowTargetTime, it will launch 1000 t2.micros.
        """
        allocation = self.run1000JobsOnMicros(jobCores=0,
                                              jobMem=0,
                                              jobDisk=0,
                                              jobTime=300,
                                              globalTargetTime=0)
        self.assertEqual(allocation, {t2_micro: 1})

    def testLongRunningJobs(self):
        """
        Test that jobs with long run times (especially service jobs) are aggressively parallelized.

        This is important, because services are one case where the degree of parallelization
        really, really matters. If you have multiple services, they may all need to be running
        simultaneously before any real work can be done.

        Despite setting globalTargetTime=3600, this should launch 1000 t2.micros because each job's
        estimated runtime (30000 seconds) extends well beyond 3600 seconds.
        """
        allocation = self.run1000JobsOnMicros(jobCores=1,
                                              jobMem=h2b('1G'),
                                              jobDisk=h2b('1G'),
                                              jobTime=30000,
                                              globalTargetTime=3600)
        self.assertEqual(allocation, {t2_micro: 1000})

    def run1000JobsOnMicros(self, jobCores, jobMem, jobDisk, jobTime, globalTargetTime):
        """Test packing 1000 jobs on t2.micros.  Depending on the targetTime and resources,
        these should pack differently.
        """
        node_shapes_for_testing = [t2_micro]
        bpf = BinPackedFit(node_shapes_for_testing, targetTime=globalTargetTime)

        for _ in range(1000):
            bpf.addJobShape(Shape(wallTime=jobTime,
                                   memory=jobMem,
                                   cores=jobCores,
                                   disk=jobDisk,
                                   preemptible=False))
        return bpf.getRequiredNodes()

    def testPathologicalCase(self):
        """Test a pathological case where only one node can be requested to fit months' worth of jobs.

        If the reservation is extended to fit a long job, and the
        bin-packer naively searches through all the reservation slices
        to find the first slice that fits, it will happily assign the
        first slot that fits the job, even if that slot occurs days in
        the future.
        """
        # Add one job that partially fills an r3.8xlarge for 1000 hours
        self.bpf.addJobShape(Shape(wallTime=3600000,
                                   memory=h2b('10G'),
                                   cores=0,
                                   disk=h2b('10G'),
                                   preemptible=False))
        for _ in range(500):
            # Add 500 CPU-hours worth of jobs that fill an r3.8xlarge
            self.bpf.addJobShape(Shape(wallTime=3600,
                                       memory=h2b('26G'),
                                       cores=32,
                                       disk=h2b('60G'),
                                       preemptible=False))
        # Hopefully we didn't assign just one node to cover all those jobs.
        self.assertNotEqual(self.bpf.getRequiredNodes(), {r3_8xlarge: 1, c4_8xlarge_preemptible: 0})

    def testJobTooLargeForAllNodes(self):
        """
        If a job is too large for all node types, the scaler should print a
        warning, but definitely not crash.
        """
        # Takes more RAM than an r3.8xlarge
        largerThanR3 = Shape(wallTime=3600,
                             memory=h2b('360G'),
                             cores=32,
                             disk=h2b('600G'),
                             preemptible=False)
        self.bpf.addJobShape(largerThanR3)
        # If we got here we didn't crash.

class ClusterScalerTest(ToilTest):
    def setUp(self):
        super().setUp()
        self.config = Config()
        self.config.targetTime = 1800
        self.config.nodeTypes = [r3_8xlarge, c4_8xlarge_preemptible]

        # Set up the mock leader
        self.leader = MockBatchSystemAndProvisioner(config=self.config, secondsPerJob=1)
        # It is also a full mock provisioner, so configure it to be that as well
        self.provisioner = self.leader
        # Pretend that Shapes are actually strings we can use for instance type names.
        self.provisioner.setAutoscaledNodeTypes([({t}, None) for t in self.config.nodeTypes])

    def testRounding(self):
        """
        Test to make sure the ClusterScaler's rounding rounds properly.
        """

        # Get a ClusterScaler
        self.config.targetTime = 1
        self.config.betaInertia = 0.0
        self.config.maxNodes = [2, 3]
        scaler = ClusterScaler(self.provisioner, self.leader, self.config)

        # Exact integers round to themselves
        self.assertEqual(scaler._round(0.0), 0)
        self.assertEqual(scaler._round(1.0), 1)
        self.assertEqual(scaler._round(-1.0), -1)
        self.assertEqual(scaler._round(123456789101112.13), 123456789101112)

        # Decimals other than X.5 round to the side they are closer to
        self.assertEqual(scaler._round(1E-10), 0)
        self.assertEqual(scaler._round(0.5 + 1E-15), 1)
        self.assertEqual(scaler._round(-0.9), -1)
        self.assertEqual(scaler._round(-0.4), 0)

        # Decimals at exactly X.5 round away from 0
        self.assertEqual(scaler._round(0.5), 1)
        self.assertEqual(scaler._round(-0.5), -1)
        self.assertEqual(scaler._round(2.5), 3)
        self.assertEqual(scaler._round(-2.5), -3)
        self.assertEqual(scaler._round(15.5), 16)
        self.assertEqual(scaler._round(-15.5), -16)
        self.assertEqual(scaler._round(123456789101112.5), 123456789101113)

    def testMaxNodes(self):
        """
        Set the scaler to be very aggressive, give it a ton of jobs, and
        make sure it doesn't go over maxNodes.
        """
        self.config.targetTime = 1
        self.config.betaInertia = 0.0
        self.config.maxNodes = [2, 3]
        scaler = ClusterScaler(self.provisioner, self.leader, self.config)
        jobShapes = [Shape(wallTime=3600,
                           cores=2,
                           memory=h2b('1G'),
                           disk=h2b('2G'),
                           preemptible=True)] * 1000
        jobShapes.extend([Shape(wallTime=3600,
                                cores=2,
                                memory=h2b('1G'),
                                disk=h2b('2G'),
                                preemptible=False)] * 1000)
        estimatedNodeCounts, could_not_fit = scaler.getEstimatedNodeCounts(jobShapes, defaultdict(int))
        self.assertEqual(estimatedNodeCounts[r3_8xlarge], 2)
        self.assertEqual(estimatedNodeCounts[c4_8xlarge_preemptible], 3)
        self.assertEqual(len(could_not_fit), 0)

    def testMinNodes(self):
        """
        Without any jobs queued, the scaler should still estimate "minNodes" nodes.
        """
        self.config.betaInertia = 0.0
        self.config.minNodes = [2, 3]
        scaler = ClusterScaler(self.provisioner, self.leader, self.config)
        jobShapes = []
        estimatedNodeCounts, could_not_fit = scaler.getEstimatedNodeCounts(jobShapes, defaultdict(int))
        self.assertEqual(estimatedNodeCounts[r3_8xlarge], 2)
        self.assertEqual(estimatedNodeCounts[c4_8xlarge_preemptible], 3)
        self.assertEqual(len(could_not_fit), 0)

    def testPreemptibleDeficitResponse(self):
        """
        When a preemptible deficit was detected by a previous run of the
        loop, the scaler should add non-preemptible nodes to
        compensate in proportion to preemptibleCompensation.
        """
        self.config.targetTime = 1
        self.config.betaInertia = 0.0
        self.config.maxNodes = [10, 10]
        # This should mean that one non-preemptible node is launched
        # for every two preemptible nodes "missing".
        self.config.preemptibleCompensation = 0.5
        # In this case, we want to explicitly set up the config so
        # that we can have preemptible and non-preemptible nodes of
        # the same type. That is the only situation where
        # preemptibleCompensation applies.
        self.config.nodeTypes = [c4_8xlarge_preemptible, c4_8xlarge]
        self.provisioner.setAutoscaledNodeTypes([({t}, None) for t in self.config.nodeTypes])

        scaler = ClusterScaler(self.provisioner, self.leader, self.config)
        # Simulate a situation where a previous run caused a
        # "deficit" of 5 preemptible nodes (e.g. a spot bid was lost)
        scaler.preemptibleNodeDeficit[c4_8xlarge] = 5
        # Add a bunch of preemptible jobs (so the bin-packing
        # estimate for the non-preemptible node should still be 0)
        jobShapes = [Shape(wallTime=3600,
                           cores=2,
                           memory=h2b('1G'),
                           disk=h2b('2G'),
                           preemptible=True)] * 1000
        estimatedNodeCounts, could_not_fit = scaler.getEstimatedNodeCounts(jobShapes, defaultdict(int))
        # We don't care about the estimated size of the preemptible
        # nodes. All we want to know is if we responded to the deficit
        # properly: 0.5 * 5 (preemptibleCompensation * the deficit) = 3 (rounded up).
        self.assertEqual(estimatedNodeCounts[self.provisioner.node_shapes_for_testing[1]], 3)
        self.assertEqual(len(could_not_fit), 0)

    def testPreemptibleDeficitIsSet(self):
        """
        Make sure that updateClusterSize sets the preemptible deficit if
        it can't launch preemptible nodes properly. That way, the
        deficit can be communicated to the next run of
        estimateNodeCount.
        """
        # Mock out addNodes. We want to pretend it had trouble
        # launching all 5 nodes, and could only launch 3.
        self.provisioner.addNodes = MagicMock(return_value=3)
        # Pretend there are no nodes in the cluster right now
        self.provisioner.getProvisionedWorkers = MagicMock(return_value=[])
        # In this case, we want to explicitly set up the config so
        # that we can have preemptible and non-preemptible nodes of
        # the same type. That is the only situation where
        # preemptibleCompensation applies.
        self.config.nodeTypes = [c4_8xlarge_preemptible, c4_8xlarge]
        self.provisioner.setAutoscaledNodeTypes([({t}, None) for t in self.config.nodeTypes])
        scaler = ClusterScaler(self.provisioner, self.leader, self.config)
        estimatedNodeCounts = {c4_8xlarge_preemptible: 5, c4_8xlarge: 0}
        scaler.updateClusterSize(estimatedNodeCounts)
        self.assertEqual(scaler.preemptibleNodeDeficit[c4_8xlarge_preemptible], 2)
        self.provisioner.addNodes.assert_called_once()

        # OK, now pretend this is a while later, and actually launched
        # the nodes properly. The deficit should disappear
        self.provisioner.addNodes = MagicMock(return_value=5)
        scaler.updateClusterSize(estimatedNodeCounts)
        self.assertEqual(scaler.preemptibleNodeDeficit[c4_8xlarge], 0)

    def testNoLaunchingIfDeltaAlreadyMet(self):
        """
        Check that the scaler doesn't try to launch "0" more instances if
        the delta was able to be met by unignoring nodes.
        """
        # We have only one node type for simplicity
        self.provisioner.setAutoscaledNodeTypes([({c4_8xlarge}, None)])
        scaler = ClusterScaler(self.provisioner, self.leader, self.config)
        # Pretend there is one ignored worker in the cluster
        self.provisioner.getProvisionedWorkers = MagicMock(
            return_value=[Node('127.0.0.1', '127.0.0.1', 'testNode',
                               datetime.datetime.now().isoformat(),
                               nodeType=c4_8xlarge, preemptible=True)])
        scaler.ignoredNodes.add('127.0.0.1')
        # Exercise the updateClusterSize logic
        self.provisioner.addNodes = MagicMock()
        scaler.updateClusterSize({c4_8xlarge: 1})
        self.assertFalse(self.provisioner.addNodes.called,
                         "addNodes was called when no new nodes were needed")
        self.assertEqual(len(scaler.ignoredNodes), 0,
                         "The scaler didn't unignore an ignored node when "
                         "scaling up")

    def testBetaInertia(self):
        # This is really high, but makes things easy to calculate.
        self.config.betaInertia = 0.5
        scaler = ClusterScaler(self.provisioner, self.leader, self.config)
        # OK, smoothing things this much should get us 50% of the way to 100.
        self.assertEqual(scaler.smoothEstimate(c4_8xlarge_preemptible, 100), 50)
        # Now we should be at 75%.
        self.assertEqual(scaler.smoothEstimate(c4_8xlarge_preemptible, 100), 75)
        # We should eventually converge on our estimate as long as betaInertia is below 1.
        for _ in range(1000):
            scaler.smoothEstimate(c4_8xlarge_preemptible, 100)
        self.assertEqual(scaler.smoothEstimate(c4_8xlarge_preemptible, 100), 100)

    def test_overhead_accounting_large(self):
        """
        If a node has a certain raw memory or disk capacity, that won't all be
        available when it actually comes up; some disk and memory will be used
        by the OS, and the backing scheduler (Mesos, Kubernetes, etc.).

        Make sure this overhead is accounted for for large nodes.
        """

        # Use a small node (60G) and a big node (260G)

        # If the job needs 100% of the memory of the instance type, it won't
        # fit and will need a bigger node.
        self._check_job_estimate([(c4_8xlarge, 0), (r3_8xlarge, 1)], memory=h2b('60G'))

        # If the job needs 98% of the memory of the instance type, it won't
        # fit and will need a bigger node.
        self._check_job_estimate([(c4_8xlarge, 0), (r3_8xlarge, 1)], memory=int(h2b('60G') * 0.98))

        # If the job needs 90% of the memory of the instance type, it will fit.
        self._check_job_estimate([(c4_8xlarge, 1), (r3_8xlarge, 0)], memory=int(h2b('60G') * 0.90))

        # If the job needs 100% of the disk of the instance type, it won't
        # fit and will need a bigger node.
        self._check_job_estimate([(c4_8xlarge, 0), (r3_8xlarge, 1)], disk=h2b('100G'))

        # If the job needs all but 7G of the disk of the instance type, it won't
        # fit and will need a bigger node.
        self._check_job_estimate([(c4_8xlarge, 0), (r3_8xlarge, 1)], disk=h2b('93G'))

        # If the job leaves 10% and 10G of the disk free, it fits
        self._check_job_estimate([(c4_8xlarge, 1), (r3_8xlarge, 0)], disk=h2b('90G'))

    def test_overhead_accounting_small(self):
        """
        If a node has a certain raw memory or disk capacity, that won't all be
        available when it actually comes up; some disk and memory will be used
        by the OS, and the backing scheduler (Mesos, Kubernetes, etc.).

        Make sure this overhead is accounted for for small nodes.
        """

        # Use a small node (1G) and a big node (260G)

        # If the job needs 100% of the memory of the instance type, it won't
        # fit and will need a bigger node.
        self._check_job_estimate([(t2_micro, 0), (r3_8xlarge, 1)], memory=h2b('1G'))

        # If the job needs all but 100M of the memory of the instance type, it
        # won't fit and will need a bigger node.
        self._check_job_estimate([(t2_micro, 0), (r3_8xlarge, 1)], memory=h2b('1G') - h2b('100M'))

        # If the job needs no more than 90% of the memory on the node *and*
        # leaves at least 384M free for overhead, we can rely on it fitting on a 1G
        # memory node.
        jobs = [
            Shape(
                wallTime=3600,
                cores=1,
                memory=h2b('1G') - h2b('384M'),
                disk=h2b('2G'),
                preemptible=True
            )
        ]
        self._check_job_estimate([(t2_micro, 1), (r3_8xlarge, 0)], memory=h2b('1G') - h2b('384M'))

    def test_overhead_accounting_observed(self):
        """
        If a node has a certain raw memory or disk capacity, that won't all be
        available when it actually comes up; some disk and memory will be used
        by the OS, and the backing scheduler (Mesos, Kubernetes, etc.).

        Make sure this overhead is accounted for so that real-world observed
        failures cannot happen again.
        """
        # In
        # https://github.com/DataBiosphere/toil/issues/4147#issuecomment-1179587561
        # a user observed what seems to be a nominally 64 GiB node
        # (r5.2xlarge?) with Mesos reporting "61.0 GB" available, and a
        # nominally 32 GiB node with Mesos reporting "29.9 GB" available. It's
        # not clear if Mesos is thinking in actual GB or GiB here.

        # A 62.5Gi job is sent to the larger node
        self._check_job_estimate([(r5_2xlarge, 0), (r5_4xlarge, 1)], memory=h2b('62.5 Gi'))

    def _check_job_estimate(self, nodes: List[Tuple[Shape, int]], cores=1, memory=1, disk=1) -> None:
        """
        Make sure that a job with the given requirements, when run on the given
        nodes, produces the given numbers of them.
        """

        self.provisioner.setAutoscaledNodeTypes([({node}, None) for node, _ in nodes])
        self.config.targetTime = 1
        self.config.betaInertia = 0.0
        # We only need up to one node
        self.config.maxNodes = [1] * len(nodes)
        scaler = ClusterScaler(self.provisioner, self.leader, self.config)

        jobs = [
            Shape(
                wallTime=3600,
                cores=cores,
                memory=memory,
                disk=disk,
                preemptible=True
            )
        ]

        logger.debug('Try and fit jobs: %s', jobs)
        counts, could_not_fit = scaler.getEstimatedNodeCounts(jobs, defaultdict(int))
        for node, count in nodes:
            seen_count = counts.get(node, 0)
            if seen_count != count:
                logger.error('Saw %s/%s instances of node %s', seen_count, count, node)
            self.assertEqual(seen_count, count)
        self.assertEqual(len(could_not_fit), 0)

class ScalerThreadTest(ToilTest):
    def _testClusterScaling(self, config, numJobs, numPreemptibleJobs, jobShape):
        """
        Test the ClusterScaler class with different patterns of job creation. Tests ascertain that
        autoscaling occurs and that all the jobs are run.
        """
        # First do simple test of creating 100 preemptible and non-premptable jobs and check the
        # jobs are completed okay, then print the amount of worker time expended and the total
        # number of worker nodes used.

        mock = MockBatchSystemAndProvisioner(config=config, secondsPerJob=2.0)
        mock.setAutoscaledNodeTypes([({t}, None) for t in config.nodeTypes])
        mock.start()
        clusterScaler = ScalerThread(mock, mock, config, stop_on_exception=True)
        clusterScaler.start()
        try:
            # Add 100 jobs to complete
            list(map(lambda x: mock.addJob(jobShape=jobShape),
                     list(range(numJobs))))
            list(map(lambda x: mock.addJob(jobShape=jobShape, preemptible=True),
                     list(range(numPreemptibleJobs))))

            # Add some completed jobs
            for preemptible in (True, False):
                if preemptible and numPreemptibleJobs > 0 or not preemptible and numJobs > 0:
                    # Add 1000 random jobs
                    for _ in range(1000):
                        x = mock.getNodeShape(nodeType=jobShape)
                        iJ = JobDescription(requirements=dict(
                                                memory=random.randrange(1, x.memory),
                                                cores=random.randrange(1, x.cores),
                                                disk=random.randrange(1, x.disk),
                                                preemptible=preemptible),
                                            jobName='testClusterScaling', unitName='')
                        clusterScaler.addCompletedJob(iJ, random.choice(list(range(1, x.wallTime))))

            startTime = time.time()
            # Wait while the cluster processes the jobs
            while (mock.getNumberOfJobsIssued(preemptible=False) > 0
                   or mock.getNumberOfJobsIssued(preemptible=True) > 0
                   or mock.getNumberOfNodes() > 0 or mock.getNumberOfNodes(preemptible=True) > 0):
                logger.debug("Running, non-preemptible queue size: %s, non-preemptible workers: %s, "
                            "preemptible queue size: %s, preemptible workers: %s" %
                            (mock.getNumberOfJobsIssued(preemptible=False),
                             mock.getNumberOfNodes(preemptible=False),
                             mock.getNumberOfJobsIssued(preemptible=True),
                             mock.getNumberOfNodes(preemptible=True)))
                clusterScaler.check()
                time.sleep(0.5)
            logger.debug("We waited %s for cluster to finish" % (time.time() - startTime))
        finally:
            clusterScaler.shutdown()
            mock.shutDown()

        # Print some info about the autoscaling
        logger.debug("Total-jobs: %s: Max-workers: %s, "
                     "Total-worker-time: %s, Worker-time-per-job: %s" %
                    (mock.totalJobs, sum(mock.maxWorkers.values()),
                     mock.totalWorkerTime,
                     mock.totalWorkerTime // mock.totalJobs if mock.totalJobs > 0 else 0.0))

    @slow
    def testClusterScaling(self):
        """
        Test scaling for a batch of non-preemptible jobs and no preemptible jobs (makes debugging
        easier).
        """
        config = Config()

        # Make defaults dummy values
        config.defaultMemory = h2b('1Gi')
        config.defaultCores = 1
        config.defaultDisk = h2b('1Gi')

        # No preemptible nodes/jobs
        config.maxPreemptibleNodes = []  # No preemptible nodes

        # Non-preemptible parameters
        config.nodeTypes = [Shape(20, h2b('10Gi'), 10, h2b('100Gi'), False)]
        config.minNodes = [0]
        config.maxNodes = [10]

        # Algorithm parameters
        config.targetTime = defaultTargetTime
        config.betaInertia = 0.1
        config.scaleInterval = 3

        self._testClusterScaling(config, numJobs=100, numPreemptibleJobs=0,
                                 jobShape=Shape(20, h2b('7Gi'), 10, h2b('80Gi'), False))

    @slow
    def testClusterScalingMultipleNodeTypes(self):

        small_node = Shape(20, h2b('5Gi'), 10, h2b('20Gi'), False)
        small_job = Shape(20, h2b('3Gi'), 10, h2b('4Gi'), False)
        medium_node = Shape(20, h2b('10Gi'), 10, h2b('20Gi'), False)
        medium_job = Shape(20, h2b('7Gi'), 10, h2b('4Gi'), False)
        large_node = Shape(20, h2b('20Gi'), 10, h2b('20Gi'), False)
        large_job = Shape(20, h2b('16Gi'), 10, h2b('4Gi'), False)

        numJobs = 100

        config = Config()

        # Make defaults dummy values
        config.defaultMemory = h2b('1Gi')
        config.defaultCores = 1
        config.defaultDisk = h2b('1Gi')

        # No preemptible nodes/jobs
        config.preemptibleNodeTypes = []
        config.minPreemptibleNodes = []
        config.maxPreemptibleNodes = []  # No preemptible nodes

        # Make sure the node types don't have to be ordered
        config.nodeTypes = [large_node, small_node, medium_node]
        config.minNodes = [0, 0, 0]
        config.maxNodes = [10, 10]  # test expansion of this list

        # Algorithm parameters
        config.targetTime = defaultTargetTime
        config.betaInertia = 0.1
        config.scaleInterval = 3

        mock = MockBatchSystemAndProvisioner(config=config, secondsPerJob=2.0)
        mock.setAutoscaledNodeTypes([({t}, None) for t in config.nodeTypes])
        clusterScaler = ScalerThread(mock, mock, config, stop_on_exception=True)
        clusterScaler.start()
        mock.start()

        try:
            # Add small jobs
            list(map(lambda x: mock.addJob(jobShape=small_job), list(range(numJobs))))
            list(map(lambda x: mock.addJob(jobShape=medium_job), list(range(numJobs))))

            # Add medium completed jobs
            for i in range(1000):
                iJ = JobDescription(requirements=dict(
                                        memory=random.choice(range(small_job.memory, medium_job.memory)),
                                        cores=medium_job.cores,
                                        disk=large_job.disk,
                                        preemptible=False),
                                    jobName='testClusterScaling', unitName='')
                clusterScaler.addCompletedJob(iJ, random.choice(range(1, 10)))

            while mock.getNumberOfJobsIssued() > 0 or mock.getNumberOfNodes() > 0:
                logger.debug("%i nodes currently provisioned" % mock.getNumberOfNodes())
                # Make sure there are no large nodes
                self.assertEqual(mock.getNumberOfNodes(nodeType=large_node), 0)
                clusterScaler.check()
                time.sleep(0.5)
        finally:
            clusterScaler.shutdown()
            mock.shutDown()

        # Make sure jobs ran on both the small and medium node types
        self.assertTrue(mock.totalJobs > 0)
        self.assertTrue(mock.maxWorkers[small_node] > 0)
        self.assertTrue(mock.maxWorkers[medium_node] > 0)

        self.assertEqual(mock.maxWorkers[large_node], 0)

    @slow
    def testClusterScalingWithPreemptibleJobs(self):
        """
        Test scaling simultaneously for a batch of preemptible and non-preemptible jobs.
        """
        config = Config()

        node_shape = Shape(20, h2b('10Gi'), 10, h2b('20Gi'), False)
        preemptible_node_shape = Shape(20, h2b('10Gi'), 10, h2b('20Gi'), True)
        job_shape = Shape(20, h2b('7Gi'), 10, h2b('2Gi'), False)
        preemptible_job_shape = Shape(20, h2b('7Gi'), 10, h2b('2Gi'), True)

        # Make defaults dummy values
        config.defaultMemory = h2b('1Gi')
        config.defaultCores = 1
        config.defaultDisk = h2b('1Gi')

        # non-preemptible node parameters
        config.nodeTypes = [node_shape, preemptible_node_shape]
        config.minNodes = [0, 0]
        config.maxNodes = [10, 10]

        # Algorithm parameters
        config.targetTime = defaultTargetTime
        config.betaInertia = 0.9
        config.scaleInterval = 3

        self._testClusterScaling(config, numJobs=100, numPreemptibleJobs=100, jobShape=job_shape)


class MockBatchSystemAndProvisioner(AbstractScalableBatchSystem, AbstractProvisioner):
    """Mimics a leader, job batcher, provisioner and scalable batch system."""
    def __init__(self, config, secondsPerJob):
        super().__init__(clusterName='clusterName', clusterType='mesos')
        # To mimic parallel preemptible and non-preemptible queues
        # for jobs we create two parallel instances of the following class
        self.config = config
        self.secondsPerJob = secondsPerJob
        self.provisioner = self
        self.batchSystem = self
        self.jobQueue = Queue()
        self.updatedJobsQueue = Queue()
        self.jobBatchSystemIDToIssuedJob = {}
        self.totalJobs = 0  # Count of total jobs processed
        self.totalWorkerTime = 0.0  # Total time spent in worker threads
        self.toilMetrics = None
        self.nodesToWorker = {}  # Map from Node to instances of the Worker class
        self.workers = defaultdict(list)  # Instances of the Worker class, by node shape
        self.maxWorkers = defaultdict(int)  # Maximum number of workers, by node shape
        self.running = False
        self.leaderThread = Thread(target=self._leaderFn)
        self.toilState = Namespace()
        self.toilState.bus = Namespace()
        self.toilState.bus.publish = MagicMock()

    def start(self):
        self.running = True
        self.leaderThread.start()

    def shutDown(self):
        self.running = False
        self.leaderThread.join()

    # Stub out all AbstractBatchSystem methods since they are never called
    for name, value in AbstractBatchSystem.__dict__.items():
        if getattr(value, '__isabstractmethod__', False):
            exec('def %s(): pass' % name)
        # Without this, the class would end up with .name and .value attributes
        del name, value

    # AbstractScalableBatchSystem methods
    def nodeInUse(self, nodeIP):
        return False

    def ignoreNode(self, nodeAddress):
        pass

    def unignoreNode(self, nodeAddress):
        pass

    def supportedClusterTypes(self):
        return {'mesos'}

    def createClusterSettings(self):
        pass

    def readClusterSettings(self):
        pass

    # AbstractProvisioner methods
    def setAutoscaledNodeTypes(self, node_types: List[Tuple[Set[Shape], Optional[float]]]):
        self.node_shapes_for_testing = sorted(it for t in node_types for it in t[0])
        super().setAutoscaledNodeTypes(node_types)

    def getProvisionedWorkers(self, instance_type=None, preemptible=None):
        """
        Returns a list of Node objects, each representing a worker node in the cluster

        :param preemptible: If True only return preemptible nodes else return non-preemptible nodes
        :return: list of Node
        """
        nodesToWorker = self.nodesToWorker
        results = []
        if instance_type:
            results = [node for node in nodesToWorker if node.nodeType == instance_type]
        else:
            results = list(nodesToWorker.keys())
        if preemptible is not None:
            results = [node for node in results if node.preemptible == preemptible]
        return results

    def terminateNodes(self, nodes):
        if nodes:
            self._removeNodes(nodes)

    def remainingBillingInterval(self, node):
        pass

    def addJob(self, jobShape, preemptible=False):
        """
        Add a job to the job queue
        """
        self.totalJobs += 1
        jobID = uuid.uuid4()
        self.jobBatchSystemIDToIssuedJob[jobID] = JobDescription(requirements={"memory": jobShape.memory,
                                                                               "cores": jobShape.cores,
                                                                               "disk": jobShape.disk,
                                                                               "preemptible": preemptible},
                                                                 jobName=f'job{self.totalJobs}')
        self.jobQueue.put(jobID)

    # JobBatcher functionality
    def getNumberOfJobsIssued(self, preemptible=None):
        if preemptible is not None:
            jobList = [job for job in list(self.jobQueue.queue) if
                       self.jobBatchSystemIDToIssuedJob[job].preemptible == preemptible]
            return len(jobList)
        else:
            return self.jobQueue.qsize()

    def getJobs(self):
        return self.jobBatchSystemIDToIssuedJob.values()

    # AbstractScalableBatchSystem functionality
    def getNodes(self, preemptible: Optional[bool] = False, timeout: int = 600):
        nodes = dict()
        for node in self.nodesToWorker:
            if node.preemptible == preemptible:
                worker = self.nodesToWorker[node]
                nodes[node.privateIP] = NodeInfo(coresTotal=0, coresUsed=0, requestedCores=1,
                                                 memoryTotal=0, memoryUsed=0, requestedMemory=1,
                                                 workers=1 if worker.busyEvent.is_set() else 0)
        return nodes

    # AbstractProvisioner functionality
    def addNodes(self, nodeTypes: Set[str], numNodes, preemptible) -> int:
        nodeType = next(iter(nodeTypes))
        self._addNodes(numNodes=numNodes, nodeType=nodeType, preemptible=preemptible)
        return self.getNumberOfNodes(nodeType=nodeType, preemptible=preemptible)

    def getNodeShape(self, nodeType, preemptible=False):
        # Assume node shapes and node types are the same thing for testing
        # TODO: this isn't really allowed by the type requirements on AbstractProvisioner
        return nodeType

    def getWorkersInCluster(self, nodeShape):
        return self.workers[nodeShape]

    def launchCluster(self, leaderNodeType, keyName, userTags=None,
                      vpcSubnet=None, leaderStorage=50, nodeStorage=50, botoPath=None, **kwargs):
        pass

    def destroyCluster(self) -> None:
        pass

    def getLeader(self):
        pass


    def _leaderFn(self):
        while self.running:
            updatedJobID = None
            try:
                updatedJobID = self.updatedJobsQueue.get(timeout=1.0)
            except Empty:
                continue
            if updatedJobID:
                del self.jobBatchSystemIDToIssuedJob[updatedJobID]
            time.sleep(0.1)

    def _addNodes(self, numNodes, nodeType: str, preemptible=False):
        nodeShape = self.getNodeShape(nodeType=nodeType, preemptible=preemptible)

        class Worker:
            def __init__(self, jobQueue, updatedJobsQueue, secondsPerJob):
                self.busyEvent = Event()
                self.stopEvent = Event()

                def workerFn():
                    while True:
                        if self.stopEvent.is_set():
                            return
                        try:
                            jobID = jobQueue.get(timeout=1.0)
                        except Empty:
                            continue
                        updatedJobsQueue.put(jobID)
                        self.busyEvent.set()
                        time.sleep(secondsPerJob)
                        self.busyEvent.clear()

                self.startTime = time.time()
                self.worker = Thread(target=workerFn)
                self.worker.start()

            def stop(self):
                self.stopEvent.set()
                self.worker.join()
                return time.time() - self.startTime

        for _ in range(numNodes):
            node = Node('127.0.0.1', uuid.uuid4(), 'testNode', datetime.datetime.now().isoformat()+'Z', nodeType=nodeType,
                    preemptible=preemptible)
            self.nodesToWorker[node] = Worker(self.jobQueue, self.updatedJobsQueue, self.secondsPerJob)
            self.workers[nodeShape].append(self.nodesToWorker[node])
        self.maxWorkers[nodeShape] = max(self.maxWorkers[nodeShape], len(self.workers[nodeShape]))

    def _removeNodes(self, nodes):
        logger.debug("Removing nodes. %s workers and %s to terminate.", len(self.nodesToWorker), len(nodes))
        for node in nodes:
            try:
                nodeShape = self.getNodeShape(node.nodeType, node.preemptible)
                worker = self.nodesToWorker.pop(node)
                self.workers[nodeShape].pop()
                self.totalWorkerTime += worker.stop()
            except KeyError:
                # Node isn't our responsibility
                pass

    def getNumberOfNodes(self, nodeType=None, preemptible=None):
        if nodeType:
            nodeShape = self.getNodeShape(nodeType=nodeType, preemptible=preemptible)
            return len(self.workers[nodeShape])
        else:
            return len(self.nodesToWorker)



import json
import os
import shutil
import subprocess
import unittest
import uuid
from typing import List, Optional

from toil.test import ToilTest
from toil.version import exactPython
from toil.wdl.wdl_functions import (WDLPair,
                                    WDLRuntimeError,
                                    ceil,
                                    cross,
                                    floor,
                                    length,
                                    read_boolean,
                                    read_float,
                                    read_int,
                                    read_json,
                                    read_lines,
                                    read_map,
                                    read_string,
                                    read_tsv,
                                    sub,
                                    transpose,
                                    wdl_zip,
                                    write_json,
                                    write_lines,
                                    write_map,
                                    write_tsv)


class WdlStandardLibraryFunctionsTest(ToilTest):
    """ A set of test cases for toil's wdl functions."""

    def setUp(self):
        """Runs anew before each test to create farm fresh temp dirs."""
        self.output_dir = os.path.join('/tmp/', 'toil-wdl-test-' + str(uuid.uuid4()))
        os.makedirs(self.output_dir)
        os.makedirs(os.path.join(self.output_dir, 'execution'))

    @classmethod
    def setUpClass(cls):
        pass

    def tearDown(self):
        """Clean up outputs."""
        if os.path.exists(self.output_dir):
            shutil.rmtree(self.output_dir)

    def _check_output(self, path, expected_result, strip=True):
        """ Compare expected_result to content from file."""
        with open(path) as f:
            result = f.read()
        if strip:
            result = result.strip()
        self.assertEqual(expected_result, result)

    def _write_temp_file(self, function_name, content):
        """ Write content to a temp file."""
        path = os.path.join(self.output_dir, f'{function_name}_{uuid.uuid4()}.tmp')
        with open(path, 'w') as f:
            f.write(content + '\n')
        return path

    def testFn_Sub(self):
        """Test the wdl built-in functional equivalent of 'sub()'."""
        # example from the WDL spec.
        chocolike = "I like chocolate when it's late"
        self.assertEqual("I love chocolate when it's late", sub(chocolike, 'like', 'love'))
        self.assertEqual("I like chocoearly when it's early", sub(chocolike, 'late', 'early'))
        self.assertEqual("I like chocolate when it's early", sub(chocolike, 'late$', 'early'))

    def testFn_Ceil(self):
        """Test the wdl built-in functional equivalent of 'ceil()', which converts
        a Float value into an Int by rounding up to the next higher integer"""
        assert ceil(1.999) == 2
        assert ceil(-1.5) == -1

    def testFn_Floor(self):
        """Test the wdl built-in functional equivalent of 'floor()', which converts
        a Float value into an Int by rounding down to the next lower integer"""
        assert floor(1.999) == 1
        assert floor(-1.5) == -2

    def testFn_ReadLines(self):
        """Test the wdl built-in functional equivalent of 'read_lines()'."""
        # trailing newlines are stripped; spaces are kept
        lines = 'line 1\nline 2\t\t\n \n\n'
        path = self._write_temp_file('read_lines', lines)
        self.assertEqual(['line 1', 'line 2\t\t', ' '], read_lines(path))

        # preceding newlines are kept
        lines = '\n\n\nline 1\nline 2\t\t\n '
        path = self._write_temp_file('read_lines', lines)
        self.assertEqual(['', '', '', 'line 1', 'line 2\t\t', ' '], read_lines(path))

    def testFn_ReadTsv(self):
        """Test the wdl built-in functional equivalent of 'read_tsv()'."""
        tsv = [['1', '2', '3'], ['4', '5', '6'], ['7', '8', '9']]
        tsv_str = '1\t2\t3\n4\t5\t6\n7\t8\t9'

        path = self._write_temp_file('read_tsv', tsv_str)
        self.assertEqual(tsv, read_tsv(path))

    def testFn_ReadJson(self):
        """Test the wdl built-in functional equivalent of 'read_json()'."""
        json_obj = {'str': 'some string', 'num': 3.14, 'bool': True, 'null': None, 'arr': ['test']}
        json_arr = ['1', '2']
        json_num = 3.14

        path = self._write_temp_file('read_json', json.dumps(json_obj))
        self.assertEqual(json_obj, read_json(path))

        path = self._write_temp_file('read_json', json.dumps(json_arr))
        self.assertEqual(json_arr, read_json(path))

        path = self._write_temp_file('read_json', json.dumps(json_num))
        self.assertEqual(json_num, read_json(path))

    def testFn_ReadMap(self):
        """Test the wdl built-in functional equivalent of 'read_map()'."""
        map_str = 'key1\tvalue1\nkey2\tvalue2'
        path = self._write_temp_file('read_map', map_str)
        self.assertEqual({'key1': 'value1', 'key2': 'value2'}, read_map(path))

        # extra lines and spaces are stripped, except spaces in keys are kept.
        map_str = '\n\n\nkey1   \tvalue1\nkey2\tvalue2   \n   \n                \t  \n'
        path = self._write_temp_file('read_map', map_str)
        self.assertEqual({'key1   ': 'value1', 'key2': 'value2'}, read_map(path))

    def testFn_ReadInt(self):
        """Test the wdl built-in functional equivalent of 'read_int()'."""
        num = 10
        path = self._write_temp_file('read_int', content=str(num))
        self.assertEqual(num, read_int(path))

        num = 10.0
        path = self._write_temp_file('read_int', content=str(num))
        self.assertRaises(ValueError, read_int, path)

        num = 10.5
        path = self._write_temp_file('read_int', content=str(num))
        self.assertRaises(ValueError, read_int, path)

    def testFn_ReadString(self):
        """Test the wdl built-in functional equivalent of 'read_string()'."""
        some_str = 'some string'
        path = self._write_temp_file('read_string', content=some_str)
        self.assertEqual(some_str, read_string(path))

        # with preceding newlines. Cromwell strips from the front and the end.
        path = self._write_temp_file('read_string', content='\n\n\n' + some_str)
        self.assertEqual(some_str, read_string(path))

        # with trailing newlines
        path = self._write_temp_file('read_string', content=some_str + '\n\n')
        self.assertEqual(some_str, read_string(path))

    def testFn_ReadFloat(self):
        """Test the wdl built-in functional equivalent of 'read_float()'."""
        num = 2.718281828459045
        path = self._write_temp_file('read_float', content=str(num))
        self.assertEqual(num, read_float(path))

    def testFn_ReadBoolean(self):
        """Test the wdl built-in functional equivalent of 'read_boolean()'."""
        for val in (True, False):
            path = self._write_temp_file('read_boolean', content=str(val))
            self.assertEqual(val, read_boolean(path))

            # upper
            path = self._write_temp_file('read_boolean', content=str(val).upper())
            self.assertEqual(val, read_boolean(path))

            # lower
            path = self._write_temp_file('read_boolean', content=str(val).lower())
            self.assertEqual(val, read_boolean(path))

    def testFn_WriteLines(self):
        """Test the wdl built-in functional equivalent of 'write_lines()'."""
        # 'line 1'          \n
        # 'line 2\t\t'      \n
        # ' '               \n
        # '\n'              \n
        path = write_lines(['line 1', 'line 2\t\t', ' ', '\n'], temp_dir=self.output_dir)
        self._check_output(path, 'line 1\nline 2\t\t\n \n\n\n', strip=False)

    def testFn_WriteTsv(self):
        """Test the wdl built-in functional equivalent of 'write_tsv()'."""
        path = write_tsv([['1', '2', '3'], ['4', '5', '6'], ['7', '8', '9']], temp_dir=self.output_dir)
        self._check_output(path, '1\t2\t3\n4\t5\t6\n7\t8\t9')

    def testFn_WriteJson(self):
        """Test the wdl built-in functional equivalent of 'write_json()'."""
        json_obj = {'str': 'some string', 'num': 3.14, 'bool': True, 'null': None, 'arr': ['test']}
        json_arr = ['1', '2']
        json_num = 3.14
        json_str = 'test string'
        json_bool = False
        json_null = None

        # Pair[Int, Pair[Int, Pair[Int, Pair[Int, Int]]]]
        json_pairs = WDLPair(1, WDLPair(2, WDLPair(3, WDLPair(4, 5))))

        path = write_json(json_obj, temp_dir=self.output_dir)
        self._check_output(path, '{"str":"some string","num":3.14,"bool":true,"null":null,"arr":["test"]}')

        path = write_json(json_arr, temp_dir=self.output_dir)
        self._check_output(path, '["1","2"]')

        path = write_json(json_num, temp_dir=self.output_dir)
        self._check_output(path, '3.14')

        path = write_json(json_str, temp_dir=self.output_dir)
        self._check_output(path, '"test string"')

        path = write_json(json_bool, temp_dir=self.output_dir)
        self._check_output(path, 'false')

        path = write_json(json_null, temp_dir=self.output_dir)
        self._check_output(path, 'null')

        path = write_json(json_pairs, temp_dir=self.output_dir)
        self._check_output(path, '{"left":1,"right":{"left":2,"right":{"left":3,"right":{"left":4,"right":5}}}}')

    def testFn_WriteMap(self):
        """Test the wdl built-in functional equivalent of 'write_map()'."""
        path = write_map({'key1': 'value1', 'key2': 'value2'}, temp_dir=self.output_dir)
        self._check_output(path, 'key1\tvalue1\nkey2\tvalue2')

    def testFn_Transpose(self):
        """Test the wdl built-in functional equivalent of 'transpose()'."""
        self.assertEqual([[0, 3], [1, 4], [2, 5]], transpose([[0, 1, 2], [3, 4, 5]]))
        self.assertEqual([[0, 1, 2], [3, 4, 5]], transpose([[0, 3], [1, 4], [2, 5]]))

        self.assertEqual([], transpose([]))
        self.assertEqual([], transpose([[]]))  # same as Cromwell
        self.assertEqual([[0]], transpose([[0]]))
        self.assertRaises(AssertionError, transpose, [[0, 1, 2], [3, 4, 5, 6]])

    def testFn_Length(self):
        """Test the WDL 'length()' built-in."""
        self.assertEqual(3, length([1, 2, 3]))
        self.assertEqual(3, length(['a', 'b', 'c']))
        self.assertEqual(0, length([]))

    def testFn_Zip(self):
        """Test the wdl built-in functional equivalent of 'zip()'."""
        left_array = [1, 2, 3]
        right_array = ['a', 'b', 'c']
        zipped = wdl_zip(left_array, right_array)
        expected_results = [WDLPair(1, 'a'), WDLPair(2, 'b'), WDLPair(3, 'c')]

        self.assertEqual(zipped, expected_results)

        # input with different size should fail.
        self.assertRaises(WDLRuntimeError, wdl_zip, [1, 2, 3], ['a', 'b'])

    def testFn_Cross(self):
        """Test the wdl built-in functional equivalent of 'cross()'."""
        left_array = [1, 2, 3]
        right_array = ['a', 'b']
        crossed = cross(left_array, right_array)
        expected_results = [WDLPair(1, 'a'), WDLPair(1, 'b'),
                            WDLPair(2, 'a'), WDLPair(2, 'b'),
                            WDLPair(3, 'a'), WDLPair(3, 'b')]

        self.assertEqual(crossed, expected_results)


class WdlWorkflowsTest(ToilTest):
    """
    A set of test cases for toil's conformance with WDL.

    All tests should include a simple wdl and json file for toil to run that checks the output.
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.program = os.path.abspath("src/toil/wdl/toilwdl.py")
        cls.test_path = os.path.abspath("src/toil/test/wdl")

    def check_function(self,
                       function_name: str,
                       cases: List[str],
                       json_file_name: Optional[str] = None,
                       expected_result: Optional[str] = None,
                       expected_exception: Optional[str] = None):
        """
        Run the given WDL workflow and check its output. The WDL workflow
        should store its output inside a 'output.txt' file that can be
        compared to `expected_result`.

        If `expected_exception` is set, this test passes only when both the
        workflow fails and that the given `expected_exception` string is
        present in standard error.
        """
        wdl_files = [os.path.abspath(f'{self.test_path}/{function_name}_{case}.wdl')
                     for case in cases]
        json_file = os.path.abspath(f'{self.test_path}/{json_file_name or function_name}.json')
        for wdl_file in wdl_files:
            with self.subTest(f'Testing: {wdl_file} {json_file}'):
                output_dir = f'/tmp/toil-wdl-test-{uuid.uuid4()}'
                os.makedirs(output_dir)

                if expected_exception is not None:
                    with self.assertRaises(subprocess.CalledProcessError) as context:
                        # use check_output() here so that the output is read before return.
                        subprocess.check_output([exactPython, self.program, wdl_file, json_file, '-o', output_dir],
                                                stderr=subprocess.PIPE)

                    stderr = context.exception.stderr
                    self.assertIsInstance(stderr, bytes)
                    self.assertIn(expected_exception, stderr.decode('utf-8'))

                elif expected_result is not None:
                    subprocess.check_call([exactPython, self.program, wdl_file, json_file, '-o', output_dir])
                    output = os.path.join(output_dir, 'output.txt')
                    with open(output) as f:
                        result = f.read().strip()
                    self.assertEqual(result, expected_result)

                else:
                    self.fail("Invalid test. Either `expected_result` or `expected_exception` must be set.")

                shutil.rmtree(output_dir)


class WdlLanguageSpecWorkflowsTest(WdlWorkflowsTest):
    """
    A set of test cases for toil's conformance with the WDL language specification:

    https://github.com/openwdl/wdl/blob/main/versions/development/SPEC.md#language-specification
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.test_path = os.path.abspath("src/toil/test/wdl/wdl_specification")

    def test_type_pair(self):
        # NOTE: these tests depend on read_lines(), write_json(), and select_first().

        expected_result = '[23,"twenty-three","a.bai",{"left":23,"right":"twenty-three"}]'
        self.check_function('type_pair', cases=['basic'], expected_result=expected_result)

        # tests if files from the pair type are correctly imported.
        # the array of three arrays consists content from:
        # 1. src/toil/test/wdl/testfiles/test_string.txt        -> 'A Whale of a Tale.'
        # 2. src/toil/test/wdl/testfiles/test_boolean.txt       -> 'true'
        # 3. src/toil/test/wdl/testfiles/test_int.txt           -> '11'
        expected_result = '[["A Whale of a Tale."],["true"],["11"]]'
        self.check_function('type_pair', cases=['with_files'], expected_result=expected_result)

    def test_v1_declaration(self):
        """
        Basic declaration example modified from the WDL 1.0 spec:

        https://github.com/openwdl/wdl/blob/main/versions/1.0/SPEC.md#declarations
        """
        expected_result = 'Hello, x!; Hello, y!'
        self.check_function('v1_spec', cases=['declaration'], expected_result=expected_result)


class WdlStandardLibraryWorkflowsTest(WdlWorkflowsTest):
    """
    A set of test cases for toil's conformance with the WDL built-in standard library:

    https://github.com/openwdl/wdl/blob/main/versions/development/SPEC.md#standard-library
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.test_path = os.path.abspath("src/toil/test/wdl/standard_library")

    def test_sub(self):
        # this workflow swaps the extension of a TSV file to CSV, with String and File inputs.
        self.check_function('sub', cases=['as_input'], expected_result='src/toil/test/wdl/test.csv')

        # NOTE: the result differs from Cromwell since we copy the file to the file store without
        # preserving the path. Cromwell would return 'src/toil/test/wdl/test.csv' instead.
        self.check_function('sub', cases=['as_input_with_file'], expected_result='test.csv')

    def test_size(self):
        self.check_function('size', cases=['as_command'], expected_result='19.0')

        # this workflow outputs the size of a 22-byte file in 'B', 'K', and 'Ki' separated with spaces.

        # NOTE: Cromwell treats the decimal and binary units (e.g.: 'K' and 'Ki') the same, which differs from
        # the spec (https://github.com/openwdl/wdl/blob/main/versions/development/SPEC.md#float-sizefile-string).
        # The correct output should be '22.0 0.022 0.021484375' not '22.0 0.021484375 0.021484375'
        self.check_function('size', cases=['as_output'], expected_result='22.0 0.022 0.021484375')

    def test_ceil(self):
        self.check_function('ceil', cases=['as_input', 'as_command'], expected_result='12')

    def test_floor(self):
        self.check_function('floor', cases=['as_input', 'as_command'], expected_result='11')

    def test_round(self):
        self.check_function('round', cases=['as_input', 'as_command'], expected_result='11')

    def test_stdout(self):
        self.check_function('stdout', cases=['as_output'], expected_result='A Whale of a Tale.')
        self.check_function('stderr', cases=['as_output'], expected_result='a journey straight to stderr')

    def test_read(self):
        """ Test the set of WDL read functions."""
        # NOTE: these tests depend on stdout() and the write_*() functions.

        self.check_function('read_lines', cases=['as_output'],
                            expected_result='line 1\n\t\tline 2 with tabs\n line 3\n\nline 5')

        self.check_function('read_tsv', cases=['as_output'],
                            expected_result='1\t2\t3\n4\t5\t6\n7\t8\t9')

        self.check_function('read_json', cases=['as_output'],
                            expected_result='{"key1":"value1","key2":"value2"}')

        self.check_function('read_map', cases=['as_output'],
                            expected_result='key1\tvalue1\nkey2\tvalue2')

        # primitives
        self.check_function('read_int', cases=['as_command'], expected_result='11')
        self.check_function('read_string', cases=['as_command'], expected_result='A Whale of a Tale.')
        self.check_function('read_float', cases=['as_command'], expected_result='11.2345')
        self.check_function('read_boolean', cases=['as_command'], expected_result='True')

    def test_write(self):
        """ Test the set of WDL write functions."""
        self.check_function('write_lines', cases=['as_command'],
                            expected_result='first\nsecond\nthird')

        self.check_function('write_tsv', cases=['as_command'],
                            expected_result='one\ttwo\tthree\nun\tdeux\ttrois')

        self.check_function('write_json', cases=['as_command'],
                            expected_result='{"key1":"value1","key2":"value2"}')

        self.check_function('write_map', cases=['as_command'],
                            expected_result='key1\tvalue1\nkey2\tvalue2')

    def test_range(self):
        # NOTE: this test depends on write_lines().
        self.check_function('range', cases=['as_input'],
                            expected_result='0\n1\n2\n3\n4\n5\n6\n7')

        self.check_function('range', cases=['as_input'],
                            json_file_name='range_0',
                            expected_result='')

        self.check_function('range', cases=['as_input'],
                            json_file_name='range_invalid',
                            expected_exception='WDLRuntimeError')

    def test_transpose(self):
        # NOTE: this test depends on write_tsv().

        # this workflow writes a transposed 2-dimensional array as a TSV file.
        self.check_function('transpose', cases=['as_input'], expected_result='0\t3\n1\t4\n2\t5')

    def test_length(self):
        self.check_function('length', cases=['as_input'], expected_result='3')

        self.check_function('length', cases=['as_input'],
                            json_file_name='length_invalid',
                            expected_exception='WDLRuntimeError')

        # length() should not work with Map[X, Y].
        self.check_function('length', cases=['as_input_with_map'],
                            expected_exception='WDLRuntimeError')

    def test_zip(self):
        self.check_function('zip', cases=['as_input'],
                            expected_result='[{"left":1,"right":"a"},{"left":2,"right":"b"},{"left":3,"right":"c"}]')

    def test_cross(self):
        self.check_function('cross', cases=['as_input'],
                            expected_result='[{"left":1,"right":"a"},{"left":1,"right":"b"},'
                                            '{"left":2,"right":"a"},{"left":2,"right":"b"},'
                                            '{"left":3,"right":"a"},{"left":3,"right":"b"}]')

    def test_as_pairs(self):
        self.check_function('as_pairs', cases=['as_input'],
                            expected_result='[{"left":"a","right":1},{"left":"b","right":2},{"left":"c","right":3}]')

    def test_as_map(self):
        self.check_function('as_map', cases=['as_input'], expected_result='{"a":1,"b":2,"c":3}')

    def test_keys(self):
        self.check_function('keys', cases=['as_input'], expected_result='["a","b","c"]')

    def test_collect_by_key(self):
        # NOTE: this result is expected according to the spec but differs from Cromwell.
        self.check_function('collect_by_key', cases=['as_input'], expected_result='{"a":[1,3],"b":[2]}')

    def test_flatten(self):
        self.check_function('flatten', cases=['as_input'], expected_result='[1,2,3,1,21,22]')


if __name__ == "__main__":
    unittest.main()



# Copyright (C) 2015-2021 Regents of the University of California
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from toil.job import Job
from toil.jobStores.abstractJobStore import NoSuchFileException
from toil.exceptions import FailedJobsException
from toil.test import ToilTest, slow


@slow
class ResumabilityTest(ToilTest):
    """
    https://github.com/BD2KGenomics/toil/issues/808
    """
    def test(self):
        """
        Tests that a toil workflow that fails once can be resumed without a NoSuchJobException.
        """
        options = Job.Runner.getDefaultOptions(self._getTestJobStorePath())
        options.logLevel = "INFO"
        options.retryCount = 0
        root = Job.wrapJobFn(parent)
        with self.assertRaises(FailedJobsException):
            # This one is intended to fail.
            Job.Runner.startToil(root, options)

        # Resume the workflow. Unfortunately, we have to check for
        # this bug using the logging output, since although the
        # NoSuchJobException causes the worker to fail, the batch
        # system code notices that the job has been deleted despite
        # the failure and avoids the failure.
        options.restart = True
        tempDir = self._createTempDir()
        options.logFile = os.path.join(tempDir, "log.txt")
        Job.Runner.startToil(root, options)
        with open(options.logFile) as f:
            logString = f.read()
            # We are looking for e.g. "Batch system is reporting that
            # the jobGraph with batch system ID: 1 and jobGraph
            # store ID: n/t/jobwbijqL failed with exit value 1"
            self.assertTrue("failed with exit value" not in logString)

def parent(job):
    """
    Set up a bunch of dummy child jobs, and a bad job that needs to be
    restarted as the follow on.
    """
    for _ in range(5):
        job.addChildJobFn(goodChild)
    job.addFollowOnJobFn(badChild)

def goodChild(job):
    """
    Does nothing.
    """
    return

def badChild(job):
    """
    Fails the first time it's run, succeeds the second time.
    """
    try:
        with job.fileStore.jobStore.read_shared_file_stream("alreadyRun") as fileHandle:
            fileHandle.read()
    except NoSuchFileException as ex:
        with job.fileStore.jobStore.write_shared_file_stream("alreadyRun", encrypted=False) as fileHandle:
            fileHandle.write(b"failed once\n")
        raise RuntimeError(f"this is an expected error: {str(ex)}")



import functools
import warnings
from typing import Any, Dict, Callable, Union, TypeVar, overload


def deprecated(new_function_name: str) -> Callable[..., Any]:
    def decorate(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def call(*args: Any, **kwargs: Any) -> Any:
            warnings.warn(f'WARNING: "{func.__name__}()" is deprecated.  Please use "{new_function_name}()" instead.',
                          DeprecationWarning)
            return func(*args, **kwargs)
        return call
    return decorate


def compat_bytes(s: Union[bytes, str]) -> str:
    return s.decode('utf-8') if isinstance(s, bytes) else s

# MyPy can't yet support the recursive type we would need to say "we go through
# any structure of dicts, tuples, lists, and sets and convert all bytes types
# in the keys and values to strings". I also can't work out how to make a
# TypeVar T that represents a generic where I could say wer go from T[bytes] ->
# T[str].
def compat_bytes_recursive(data: Any) -> Any:
    """
    Convert a tree of objects over bytes to objects over strings.
    """
    if isinstance(data, dict):
        # Keyed collection
        return type(data)((compat_bytes_recursive(i) for i in data.items()))
    elif isinstance(data, (tuple, list, set)):
        # Flat collection
        return type(data)((compat_bytes_recursive(i) for i in data))
    elif isinstance(data, bytes):
        # Leaf bytes
        return data.decode('utf-8')
    else:
        # Leaf non-bytes
        return data
